"""Generic extension components."""
import os
import os.path as op
import re
import codecs
import copy

from pyrevit import HOST_APP, PyRevitException
from pyrevit import coreutils
from pyrevit.coreutils import yaml
from pyrevit.coreutils import applocales
from pyrevit.coreutils import pyutils
from pyrevit.compat import PY3
from pyrevit.revit import ui
import pyrevit.extensions as exts


#pylint: disable=W0703,C0302,C0103
mlogger = coreutils.logger.get_logger(__name__)


EXT_DIR_KEY = 'directory'
SUB_CMP_KEY = 'components'
LAYOUT_ITEM_KEY = 'layout_items'
LAYOUT_DIR_KEY = 'directive'
TYPE_ID_KEY = 'type_id'
NAME_KEY = 'name'


class TypedComponent(object):
    """Component with a type id."""
    type_id = None


class CachableComponent(TypedComponent):
    """Cacheable Component."""
    def get_cache_data(self):
        cache_dict = self.__dict__.copy()
        if hasattr(self, TYPE_ID_KEY):
            cache_dict[TYPE_ID_KEY] = getattr(self, TYPE_ID_KEY)
        return cache_dict

    def load_cache_data(self, cache_dict):
        for k, v in cache_dict.items():
            self.__dict__[k] = v


class LayoutDirective(CachableComponent):
    """Layout directive."""
    def __init__(self, directive_type=None, target=None):
        self.directive_type = directive_type
        self.target = target


class LayoutItem(CachableComponent):
    """Layout item."""
    def __init__(self, name=None, directive=None):
        self.name = name
        self.directive = directive


class GenericComponent(CachableComponent):
    """Generic component object."""
    def __init__(self):
        self.name = None

    def __repr__(self):
        return '<GenericComponent object with name \'{}\'>'.format(self.name)

    @property
    def is_container(self):
        return hasattr(self, '__iter__')


class GenericUIComponent(GenericComponent):
    """Generic UI component."""
    def __init__(self, cmp_path=None):
        # using classname otherwise exceptions in superclasses won't show
        GenericComponent.__init__(self)
        self.directory = cmp_path
        self.unique_name = self.parent_ctrl_id = None
        self.icon_file = None
        self._ui_title = None
        self._tooltip = self.author = self._help_url = None
        self.media_file = None
        self.min_revit_ver = self.max_revit_ver = None
        self.is_beta = False
        self.highlight_type = None
        self.collapsed = False
        self.version = None

        self.meta = {}
        self.meta_file = None

        self.modules = []
        self.module_paths = []

        self.binary_path = None
        self.library_path = None

        if self.directory:
            self._update_from_directory()

    @classmethod
    def matches(cls, component_path):
        return component_path.lower().endswith(cls.type_id)

    @classmethod
    def make_unique_name(cls, cmp_path):
        """Creates a unique name for the command.

        This is used to uniquely identify this command
        and also to create the class in pyRevit dll assembly.
        Current method create a unique name based on the command
        full directory address.

        Examples:
            for 'pyRevit.extension/pyRevit.tab/Edit.panel/Flip doors.pushbutton'
            unique name would be: 'pyrevit-pyrevit-edit-flipdoors'.
        """
        pieces = []
        inside_ext = False
        for dname in cmp_path.split(op.sep):
            if exts.ExtensionTypes.UI_EXTENSION.POSTFIX in dname:
                inside_ext = True

            name, ext = op.splitext(dname)
            if ext != '' and inside_ext:
                pieces.append(name)
            else:
                continue
        return coreutils.cleanup_string(
            exts.UNIQUE_ID_SEPARATOR.join(pieces),
            skip=[exts.UNIQUE_ID_SEPARATOR]
            ).lower()

    def __repr__(self):
        return '<type_id \'{}\' name \'{}\' @ \'{}\'>'\
            .format(self.type_id, self.name, self.directory)

    def _update_from_directory(self):
        self.name = op.splitext(op.basename(self.directory))[0]
        self._ui_title = self.name
        self.unique_name = GenericUIComponent.make_unique_name(self.directory)

        self.icon_file = ui.resolve_icon_file(self.directory, exts.DEFAULT_ICON_FILE)
        mlogger.debug('Icon file is: %s:%s', self.name, self.icon_file)

        self.media_file = \
            self.find_bundle_file([exts.DEFAULT_MEDIA_FILENAME], finder='name')
        mlogger.debug('Media file is: %s:%s', self.name, self.media_file)

        self._help_url = \
            self.find_bundle_file([exts.HELP_FILE_PATTERN], finder='regex')

        # each component can store custom libraries under
        # lib/ inside the component folder
        lib_path = op.join(self.directory, exts.COMP_LIBRARY_DIR_NAME)
        self.library_path = lib_path if op.exists(lib_path) else None

        # setting up search paths. These paths will be added to
        # all sub-components of this component.
        if self.library_path:
            self.module_paths.append(self.library_path)

        # each component can store custom binaries under
        # bin/ inside the component folder
        bin_path = op.join(self.directory, exts.COMP_BIN_DIR_NAME)
        self.binary_path = bin_path if op.exists(bin_path) else None

        # setting up search paths. These paths will be added to
        # all sub-components of this component.
        if self.binary_path:
            self.module_paths.append(self.binary_path)

        # find meta file
        self.meta_file = self.find_bundle_file([
            exts.BUNDLEMATA_POSTFIX
            ])
        if self.meta_file:
            # sets up self.meta
            try:
                self.meta = yaml.load_as_dict(self.meta_file)
                if self.meta:
                    self._read_bundle_metadata()
            except Exception as err:
                mlogger.error(
                    "Error reading meta file @ %s | %s", self.meta_file, err
                    )

    def _resolve_locale(self, source):
        if isinstance(source, str):
            return source
        elif isinstance(source, dict):
            return applocales.get_locale_string(source)

    def _resolve_liquid_tag(self, param_name, key, value):
        liquid_tag = '{{' + key + '}}'
        exst_val = getattr(self, param_name)
        if exst_val and (liquid_tag in exst_val):   #pylint: disable=E1135
            new_value = exst_val.replace(liquid_tag, value)
            setattr(self, param_name, new_value)

    def _read_bundle_metadata(self):
        self._ui_title = self.meta.get(exts.MDATA_UI_TITLE, self._ui_title)

        self._tooltip = self.meta.get(exts.MDATA_TOOLTIP, self._tooltip)

        # authors could be a list or single value
        self.author = self.meta.get(exts.MDATA_AUTHOR, self.author)
        self.author = self.meta.get(exts.MDATA_AUTHORS, self.author)
        if isinstance(self.author, list):
            self.author = '\n'.join(self.author)

        self._help_url = \
            self.meta.get(exts.MDATA_COMMAND_HELP_URL, self._help_url)

        self.min_revit_ver = \
            self.meta.get(exts.MDATA_MIN_REVIT_VERSION, self.min_revit_ver)
        self.max_revit_ver = \
            self.meta.get(exts.MDATA_MAX_REVIT_VERSION, self.max_revit_ver)

        self.is_beta = \
            self.meta.get(exts.MDATA_BETA_SCRIPT, 'false').lower() == 'true'

        highlight = \
            self.meta.get(exts.MDATA_HIGHLIGHT_KEY, None)
        if highlight and isinstance(highlight, str):
            self.highlight_type = highlight.lower()

        self.collapsed = \
            self.meta.get(exts.MDATA_COLLAPSED_KEY, 'false').lower() == 'true'

        self.modules = \
            self.meta.get(exts.MDATA_LINK_BUTTON_MODULES, self.modules)

    @property
    def control_id(self):
        if self.parent_ctrl_id:
            return self.parent_ctrl_id + '%{}'.format(self.name)
        else:
            return "CustomCtrl_%CustomCtrl_%{}".format(self.name)

    @property
    def ui_title(self):
        return self._resolve_locale(self._ui_title)

    @property
    def tooltip(self):
        return self._resolve_locale(self._tooltip)

    @property
    def help_url(self):
        return self._resolve_locale(self._help_url)

    @property
    def is_supported(self):
        if self.min_revit_ver:
            # If host is older than the minimum host version, raise exception
            if int(HOST_APP.version) < int(self.min_revit_ver):
                mlogger.debug('Requires min version: %s', self.min_revit_ver)
                return False
        if self.max_revit_ver:
            # If host is newer than the max host version, raise exception
            if int(HOST_APP.version) > int(self.max_revit_ver):
                mlogger.debug('Requires max version: %s', self.max_revit_ver)
                return False
        return True

    def get_full_bundle_name(self):
        return self.name + self.type_id

    def has_module_path(self, path):
        return path in self.module_paths

    def add_module_path(self, path):
        if path and not self.has_module_path(path):
            mlogger.debug('Appending syspath: %s to %s', path, self)
            self.module_paths.append(path)

    def remove_module_path(self, path):
        if path and self.has_module_path(path):
            mlogger.debug('Removing syspath: %s from %s', path, self)
            return self.module_paths.remove(path)

    def get_bundle_file(self, file_name):
        if self.directory and file_name:
            file_addr = op.join(self.directory, file_name)
            return file_addr if op.exists(file_addr) else None

    def find_bundle_file(self, patterns, finder='postfix'):
        if self.directory:
            for bundle_file in os.listdir(self.directory):
                if 'name' == finder:
                    for file_name in patterns:
                        if op.splitext(bundle_file)[0] == file_name:
                            return op.join(self.directory, bundle_file)
                elif 'postfix' == finder:
                    for file_postfix in patterns:
                        if bundle_file.endswith(file_postfix):
                            return op.join(self.directory, bundle_file)
                elif 'regex' == finder:
                    for regex_pattern in patterns:
                        if re.match(regex_pattern, bundle_file):
                            return op.join(self.directory, bundle_file)
        return None

    def find_bundle_module(self, module, by_host=False):
        # test of file_name is an actually path to a file
        if op.isfile(module):
            return module

        def build_assm_filename(module_filename):
            # build assembly by host version (assm_file_2020.ext)
            assm_name, assm_ext = op.splitext(module_filename)
            return assm_name + '_' + HOST_APP.version + assm_ext

        if by_host:
            module = build_assm_filename(module)

        # test if module is inside search paths
        for module_path in self.module_paths:
            possible_module_path = op.join(module_path, module)
            if op.isfile(possible_module_path):
                return possible_module_path

    def configure(self, config_dict):
        configurable_params = \
            ['_ui_title', '_tooltip', '_help_url', 'author']
        # get root key:value pairs
        for key, value in config_dict.items():
            for param_name in configurable_params:
                self._resolve_liquid_tag(param_name, key, value)
        # get key:value pairs grouped under special key, if exists
        templates = config_dict.get(exts.MDATA_TEMPLATES_KEY, {})
        for key, value in templates.items():
            for param_name in configurable_params:
                self._resolve_liquid_tag(param_name, key, value)


class GenericUIContainer(GenericUIComponent):
    """Superclass for all UI group items (tab, panel, button groups, stacks)."""
    allowed_sub_cmps = []

    def __init__(self, cmp_path=None):
        self.layout_items = []
        self.components = []
        # using classname otherwise exceptions in superclasses won't show
        GenericUIComponent.__init__(self, cmp_path=cmp_path)

    def _update_from_directory(self):
        # using classname otherwise exceptions in superclasses won't show
        GenericUIComponent._update_from_directory(self)
        # process layout
        # default is layout in metadata, the older layout file is deprecate
        # and is for fallback only
        if not self.parse_layout_metadata():
            mlogger.debug('Container does not have layout file defined: %s',
                self)


    def _apply_layout_directive(self, directive, component):
        # if matching directive found, process the directive
        if directive.directive_type == 'title':
            component._ui_title = directive.target

    def __iter__(self):
        # if item is not listed in layout, it will not be created
        if self.layout_items:
            mlogger.debug('Reordering components per layout file...')
            laidout_cmps = []
            for litem in self.layout_items:
                matching_cmp = \
                        next((x for x in self.components
                              if x.name == litem.name), None)
                if matching_cmp:
                    # apply directives before adding to list
                    if litem.directive:
                        self._apply_layout_directive(litem.directive,
                                                     matching_cmp)
                    laidout_cmps.append(matching_cmp)

            # insert separators and slideouts per layout definition
            mlogger.debug('Adding separators and slide outs per layout...')
            last_item_index = len(self.layout_items) - 1
            for idx, litem in enumerate(self.layout_items):
                if exts.SEPARATOR_IDENTIFIER in litem.name \
                        and idx < last_item_index:
                    separator = GenericUIComponent()
                    separator.type_id = exts.SEPARATOR_IDENTIFIER
                    laidout_cmps.insert(idx, separator)
                elif exts.SLIDEOUT_IDENTIFIER in litem.name \
                        and idx < last_item_index:
                    slideout = GenericUIComponent()
                    slideout.type_id = exts.SLIDEOUT_IDENTIFIER
                    laidout_cmps.insert(idx, slideout)

            mlogger.debug('Reordered sub_component list is: %s', laidout_cmps)
            return laidout_cmps
        else:
            return self.components

    def parse_layout_directive(self, layout_line):
        parts = re.findall(r'(.+)\[(.+):(.*)\]', layout_line)
        if parts:
            source_item, directive, target_value = parts[0]
            # cleanup values
            directive = directive.lower().strip()
            target_value = target_value.strip()
            # process any escape characters in target value
            # https://stackoverflow.com/a/4020824/2350244
            target_value = target_value.encode('utf-8')
            if PY3:
                target_value = target_value.decode('unicode_escape')
            else:
                target_value = target_value.decode('string_escape')
            # create directive obj
            return source_item, LayoutDirective(directive_type=directive,
                                                target=target_value)
        return layout_line, None

    def parse_layout_item(self, layout_line):
        if layout_line:
            layout_item_name, layout_item_drctv = \
                self.parse_layout_directive(layout_line)
            return LayoutItem(name=layout_item_name,
                              directive=layout_item_drctv)

    def parse_layout_items(self, layout_lines):
        for layout_line in layout_lines:
            layout_item = self.parse_layout_item(layout_line)
            if layout_item:
                self.layout_items.append(layout_item)
        mlogger.debug('Layout is: %s', self.layout_items)

    def parse_layout_metadata(self):
        layout = self.meta.get(exts.MDATA_LAYOUT, [])
        if layout:
            self.parse_layout_items(layout)
            return True

    def contains(self, item_name):
        return any([x.name == item_name for x in self.components])

    def add_module_path(self, path):
        if path and not self.has_module_path(path):
            mlogger.debug('Appending syspath: %s to %s', path, self)
            for component in self.components:
                component.add_module_path(path)
            self.module_paths.append(path)

    def remove_module_path(self, path):
        if path and self.has_module_path(path):
            mlogger.debug('Removing syspath: %s from %s', path, self)
            for component in self.components:
                component.remove_module_path(path)
            return self.module_paths.remove(path)

    def add_component(self, comp):
        # set search paths
        for path in self.module_paths:
            comp.add_module_path(path)
        # set its own control id on the child component
        if hasattr(comp, 'parent_ctrl_id'):
            comp.parent_ctrl_id = self.control_id
        # now add to list
        self.components.append(comp)

    def find_components_of_type(self, cmp_type):
        sub_comp_list = []
        for sub_comp in self.components:
            if isinstance(sub_comp, cmp_type):
                sub_comp_list.append(sub_comp)
            elif sub_comp.is_container:
                sub_comp_list.extend(sub_comp.find_components_of_type(cmp_type))

        return sub_comp_list

    def find_layout_items(self):
        layout_items = []
        layout_items.extend(self.layout_items)
        for sub_comp in self.components:
            if sub_comp.is_container:
                layout_items.extend(sub_comp.find_layout_items())
        return layout_items

    def configure(self, config_dict):
        # update self meta
        GenericUIComponent.configure(self, config_dict=config_dict)
        # create an updated dict to pass to children
        updated_dict = copy.deepcopy(config_dict)
        updated_dict = pyutils.merge(updated_dict, self.meta)
        # replace the meta values with the expanded values
        # so children can use the expanded
        updated_dict[exts.MDATA_UI_TITLE] = self.ui_title
        updated_dict[exts.MDATA_TOOLTIP] = self.tooltip
        updated_dict[exts.MDATA_COMMAND_HELP_URL] = self.help_url
        updated_dict[exts.AUTHOR_PARAM] = self.author
        if exts.AUTHORS_PARAM in updated_dict:
            updated_dict.pop(exts.AUTHORS_PARAM)
        for component in self:
            component.configure(updated_dict)

# superclass for all single command classes (link, push button, toggle button)
# GenericUICommand is not derived from GenericUIContainer since a command
# can not contain other elements
class GenericUICommand(GenericUIComponent):
    """Superclass for all single commands.

    The information provided by these classes will be used to create a
    push button under Revit UI. However, pyRevit expands the capabilities of
    push button beyond what is provided by Revit UI. (e.g. Toggle button
    changes it's icon based on its on/off status)
    See LinkButton and ToggleButton classes.
    """
    def __init__(self, cmp_path=None, needs_script=True):
        self.needs_script = needs_script
        self.script_file = self.config_script_file = None
        self.arguments = []
        self.context = None
        self.class_name = self.avail_class_name = None
        self.requires_clean_engine = False
        self.requires_fullframe_engine = False
        self.requires_persistent_engine = False
        self.requires_mainthread_engine = False
        # engine options specific to dynamo
        self.dynamo_path = None
        # self.dynamo_path_exec = False
        self.dynamo_path_check_existing = False
        self.dynamo_force_manual_run = False
        self.dynamo_model_nodes_info = None
        # using classname otherwise exceptions in superclasses won't show
        GenericUIComponent.__init__(self, cmp_path=cmp_path)

        mlogger.debug('Maximum host version: %s', self.max_revit_ver)
        mlogger.debug('Minimum host version: %s', self.min_revit_ver)
        mlogger.debug('command tooltip: %s', self._tooltip)
        mlogger.debug('Command author: %s', self.author)
        mlogger.debug('Command help url: %s', self._help_url)

        if self.is_beta:
            mlogger.debug('Command is in beta.')

    def _update_from_directory(self):
        # using classname otherwise exceptions in superclasses won't show
        GenericUIComponent._update_from_directory(self)

        # find script file
        self.script_file = \
            self.find_bundle_file([
                exts.PYTHON_SCRIPT_POSTFIX,
                exts.CSHARP_SCRIPT_POSTFIX,
                exts.VB_SCRIPT_POSTFIX,
                exts.RUBY_SCRIPT_POSTFIX,
                exts.DYNAMO_SCRIPT_POSTFIX,
                exts.GRASSHOPPER_SCRIPT_POSTFIX,
                exts.GRASSHOPPERX_SCRIPT_POSTFIX,
                ])

        if self.needs_script and not self.script_file:
            mlogger.error('Command %s: Does not have script file.', self)

        # if python
        if self.script_language == exts.PYTHON_LANG:
            # allow python tools to load side scripts
            self.add_module_path(self.directory)
            # read the metadata from python script if not metadata file
            if not self.meta and not self.is_cpython:
                # sets up self.meta from script global variables
                self._read_bundle_metadata_from_python_script()

        # find config scripts
        self.config_script_file = \
            self.find_bundle_file([
                exts.PYTHON_CONFIG_SCRIPT_POSTFIX,
                exts.CSHARP_CONFIG_SCRIPT_POSTFIX,
                exts.VB_CONFIG_SCRIPT_POSTFIX,
                exts.RUBY_CONFIG_SCRIPT_POSTFIX,
                exts.DYNAMO_CONFIG_SCRIPT_POSTFIX,
                exts.GRASSHOPPER_CONFIG_SCRIPT_POSTFIX,
                exts.GRASSHOPPERX_CONFIG_SCRIPT_POSTFIX,
                ])

        if not self.config_script_file:
            mlogger.debug(
                'Command %s: Does not have independent config script.',
                self)
            self.config_script_file = self.script_file

    def _read_bundle_metadata(self):
        # using classname otherwise exceptions in superclasses won't show
        GenericUIComponent._read_bundle_metadata(self)
        # determine engine configs
        if exts.MDATA_ENGINE in self.meta:
            self.requires_clean_engine = \
                self.meta[exts.MDATA_ENGINE].get(
                    exts.MDATA_ENGINE_CLEAN, 'false').lower() == 'true'
            self.requires_fullframe_engine = \
                self.meta[exts.MDATA_ENGINE].get(
                    exts.MDATA_ENGINE_FULLFRAME, 'false').lower() == 'true'
            self.requires_persistent_engine = \
                self.meta[exts.MDATA_ENGINE].get(
                    exts.MDATA_ENGINE_PERSISTENT, 'false').lower() == 'true'

            # determine if engine is required to run on main thread
            # MDATA_ENGINE_MAINTHREAD is the generic option
            rme = self.meta[exts.MDATA_ENGINE].get(
                exts.MDATA_ENGINE_MAINTHREAD, 'false') == 'true'
            # MDATA_ENGINE_DYNAMO_AUTOMATE is specific naming for dynamo
            automate = self.meta[exts.MDATA_ENGINE].get(
                exts.MDATA_ENGINE_DYNAMO_AUTOMATE, 'false') == 'true'
            self.requires_mainthread_engine = rme or automate

            # process engine options specific to dynamo
            self.dynamo_path = \
                self.meta[exts.MDATA_ENGINE].get(
                    exts.MDATA_ENGINE_DYNAMO_PATH, None)
            # self.dynamo_path_exec = \
            #     self.meta[exts.MDATA_ENGINE].get(
            #         exts.MDATA_ENGINE_DYNAMO_PATH_EXEC, 'true') == 'true'
            self.dynamo_path_check_existing = \
                self.meta[exts.MDATA_ENGINE].get(
                    exts.MDATA_ENGINE_DYNAMO_PATH_CHECK_EXIST,
                    'false') == 'true'
            self.dynamo_force_manual_run = \
                self.meta[exts.MDATA_ENGINE].get(
                    exts.MDATA_ENGINE_DYNAMO_FORCE_MANUAL_RUN,
                    'false') == 'true'
            self.dynamo_model_nodes_info = \
                self.meta[exts.MDATA_ENGINE].get(
                    exts.MDATA_ENGINE_DYNAMO_MODEL_NODES_INFO, None)

        # panel buttons should be active always
        if self.type_id == exts.PANEL_PUSH_BUTTON_POSTFIX:
            self.context = self._parse_context_directives(exts.CTX_ZERODOC)
        else:
            self.context = \
                self.meta.get(exts.MDATA_COMMAND_CONTEXT, None)
            if self.context:
                self.context = self._parse_context_directives(self.context)

    def _parse_context_list(self, context):
        context_rules = []

        str_items = [x for x in context if isinstance(x, str)]
        context_rules.append(
            exts.MDATA_COMMAND_CONTEXT_RULE.format(
                rule=exts.MDATA_COMMAND_CONTEXT_ALL_SEP.join(str_items)
                )
        )

        dict_items = [x for x in context if isinstance(x, dict)]
        for ditem in dict_items:
            context_rules.extend(self._parse_context_dict(ditem))

        return context_rules

    def _parse_context_dict(self, context):
        context_rules = []
        for ctx_key, ctx_value in context.items():
            if ctx_key == exts.MDATA_COMMAND_CONTEXT_TYPE:
                context_type = (
                    exts.MDATA_COMMAND_CONTEXT_ANY_SEP
                    if ctx_value == exts.MDATA_COMMAND_CONTEXT_ANY
                    else exts.MDATA_COMMAND_CONTEXT_ALL_SEP
                )
                continue

            if isinstance(ctx_value, str):
                ctx_value = [ctx_value]

            key = ctx_key.lower()
            condition = ""
            # all
            if key == exts.MDATA_COMMAND_CONTEXT_ALL \
                    or key == exts.MDATA_COMMAND_CONTEXT_NOTALL:
                condition = exts.MDATA_COMMAND_CONTEXT_ALL_SEP

            # any
            elif key == exts.MDATA_COMMAND_CONTEXT_ANY \
                    or key == exts.MDATA_COMMAND_CONTEXT_NOTANY:
                condition = exts.MDATA_COMMAND_CONTEXT_ANY_SEP

            # except
            elif key == exts.MDATA_COMMAND_CONTEXT_EXACT \
                    or key == exts.MDATA_COMMAND_CONTEXT_NOTEXACT:
                condition = exts.MDATA_COMMAND_CONTEXT_EXACT_SEP

            context = condition.join(
                [x for x in ctx_value if isinstance(x, str)]
                )
            formatted_rule = \
                exts.MDATA_COMMAND_CONTEXT_RULE.format(rule=context)
            if key.startswith(exts.MDATA_COMMAND_CONTEXT_NOT):
                formatted_rule = "!" + formatted_rule
            context_rules.append(formatted_rule)
        return context_rules

    def _parse_context_directives(self, context):
        context_rules = []

        if isinstance(context, str):
            context_rules.append(
                exts.MDATA_COMMAND_CONTEXT_RULE.format(rule=context)
            )
        elif isinstance(context, list):
            context_rules.extend(self._parse_context_list(context))

        elif isinstance(context, dict):
            if "rule" in context:
                return context["rule"]
            context_rules.extend(self._parse_context_dict(context))

        context_type = exts.MDATA_COMMAND_CONTEXT_ALL_SEP
        return context_type.join(context_rules)

    def _read_bundle_metadata_from_python_script(self):
        try:
            # reading script file content to extract parameters
            script_content = \
                coreutils.ScriptFileParser(self.script_file)

            self._ui_title = \
                script_content.extract_param(exts.UI_TITLE_PARAM) \
                    or self._ui_title

            script_docstring = script_content.get_docstring()
            custom_docstring = \
                script_content.extract_param(exts.DOCSTRING_PARAM)
            self._tooltip = \
                custom_docstring or script_docstring or self._tooltip

            script_author = script_content.extract_param(exts.AUTHOR_PARAM)
            script_author = script_content.extract_param(exts.AUTHORS_PARAM)
            if isinstance(script_author, list):
                script_author = '\n'.join(script_author)
            self.author = script_author or self.author

            # extracting min requried Revit and pyRevit versions
            self.max_revit_ver = \
                script_content.extract_param(exts.MAX_REVIT_VERSION_PARAM) \
                    or self.max_revit_ver
            self.min_revit_ver = \
                script_content.extract_param(exts.MIN_REVIT_VERSION_PARAM) \
                    or self.min_revit_ver
            self._help_url = \
                script_content.extract_param(exts.COMMAND_HELP_URL_PARAM) \
                    or self._help_url

            self.is_beta = \
                script_content.extract_param(exts.BETA_SCRIPT_PARAM) \
                    or self.is_beta

            self.highlight_type = \
                script_content.extract_param(exts.HIGHLIGHT_SCRIPT_PARAM) \
                    or self.highlight_type

            # only True when command is specifically asking for
            # a clean engine or a fullframe engine. False if not set.
            self.requires_clean_engine = \
                script_content.extract_param(exts.CLEAN_ENGINE_SCRIPT_PARAM) \
                    or False
            self.requires_fullframe_engine = \
                script_content.extract_param(exts.FULLFRAME_ENGINE_PARAM) \
                    or False
            self.requires_persistent_engine = \
                script_content.extract_param(exts.PERSISTENT_ENGINE_PARAM) \
                    or False

            # panel buttons should be active always
            if self.type_id == exts.PANEL_PUSH_BUTTON_POSTFIX:
                self.context = self._parse_context_directives(exts.CTX_ZERODOC)
            else:
                self.context = \
                    script_content.extract_param(exts.COMMAND_CONTEXT_PARAM)
                if self.context:
                    self.context = self._parse_context_directives(self.context)

        except Exception as parse_err:
            mlogger.log_parse_except(self.script_file, parse_err)

    @property
    def script_language(self):
        if self.script_file is not None:
            if self.script_file.endswith(exts.PYTHON_SCRIPT_FILE_FORMAT):
                return exts.PYTHON_LANG
            elif self.script_file.endswith(exts.CSHARP_SCRIPT_FILE_FORMAT):
                return exts.CSHARP_LANG
            elif self.script_file.endswith(exts.VB_SCRIPT_FILE_FORMAT):
                return exts.VB_LANG
            elif self.script_file.endswith(exts.RUBY_SCRIPT_FILE_FORMAT):
                return exts.RUBY_LANG
            elif self.script_file.endswith(exts.DYNAMO_SCRIPT_FILE_FORMAT):
                return exts.DYNAMO_LANG
            elif self.script_file.endswith(
                    exts.GRASSHOPPER_SCRIPT_FILE_FORMAT) \
                    or self.script_file.endswith(
                        exts.GRASSHOPPERX_SCRIPT_FILE_FORMAT
                    ):
                return exts.GRASSHOPPER_LANG
        else:
            return None

    @property
    def control_id(self):
        if self.parent_ctrl_id:
            return self.parent_ctrl_id + '%{}'.format(self.name)
        else:
            return '%{}'.format(self.name)

    @property
    def is_cpython(self):
        if PY3:
            with open(self.script_file, 'r', encoding='utf-8') as script_f:
                return exts.CPYTHON_HASHBANG in script_f.readline()
        with open(self.script_file, 'r') as script_f:
            return exts.CPYTHON_HASHBANG in script_f.readline()

    def has_config_script(self):
        return self.config_script_file != self.script_file



class NodeLabelEditEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.TreeView.BeforeLabelEdit and System.Windows.Forms.TreeView.AfterLabelEdit events.

 

 NodeLabelEditEventArgs(node: TreeNode)

 NodeLabelEditEventArgs(node: TreeNode,label: str)
 """
 @staticmethod
 def __new__(self,node,label=None):
  """
  __new__(cls: type,node: TreeNode)

  __new__(cls: type,node: TreeNode,label: str)
  """
  pass
 CancelEdit=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether the edit has been canceled.



Get: CancelEdit(self: NodeLabelEditEventArgs) -> bool



Set: CancelEdit(self: NodeLabelEditEventArgs)=value

"""

 Label=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the new text to associate with the tree node.



Get: Label(self: NodeLabelEditEventArgs) -> str



"""

 Node=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the tree node containing the text to edit.



Get: Node(self: NodeLabelEditEventArgs) -> TreeNode



"""





# encoding: utf-8
# module System.Collections.Concurrent calls itself Concurrent
# from mscorlib,Version=4.0.0.0,Culture=neutral,PublicKeyToken=b77a5c561934e089,System,Version=4.0.0.0,Culture=neutral,PublicKeyToken=b77a5c561934e089
# by generator 1.145
""" NamespaceTracker represent a CLS namespace. """
# no imports

# functions

def Partitioner(*args,**kwargs): # real signature unknown
 """ Provides common partitioning strategies for arrays,lists,and enumerables. """
 pass
# classes

class BlockingCollection(object,IEnumerable[T],IEnumerable,ICollection,IDisposable,IReadOnlyCollection[T]):
 """
 BlockingCollection[T]()

 BlockingCollection[T](boundedCapacity: int)

 BlockingCollection[T](collection: IProducerConsumerCollection[T],boundedCapacity: int)

 BlockingCollection[T](collection: IProducerConsumerCollection[T])
 """
 def Add(self,item,cancellationToken=None):
  """
  Add(self: BlockingCollection[T],item: T,cancellationToken: CancellationToken)

   Adds the item to the System.Collections.Concurrent.BlockingCollection.

  

   item: The item to be added to the collection. The value can be a null reference.

   cancellationToken: A cancellation token to observe.

  Add(self: BlockingCollection[T],item: T)

   Adds the item to the System.Collections.Concurrent.BlockingCollection.

  

   item: The item to be added to the collection. The value can be a null reference.
  """
  pass
 @staticmethod
 def AddToAny(collections,item,cancellationToken=None):
  """
  AddToAny(collections: Array[BlockingCollection[T]],item: T,cancellationToken: CancellationToken) -> int

  AddToAny(collections: Array[BlockingCollection[T]],item: T) -> int
  """
  pass
 def CompleteAdding(self):
  """
  CompleteAdding(self: BlockingCollection[T])

   Marks the System.Collections.Concurrent.BlockingCollection instances as not accepting any more 

    additions.
  """
  pass
 def CopyTo(self,array,index):
  """ CopyTo(self: BlockingCollection[T],array: Array[T],index: int) """
  pass
 def Dispose(self):
  """
  Dispose(self: BlockingCollection[T])

   Releases all resources used by the current instance of the 

    System.Collections.Concurrent.BlockingCollection class.
  """
  pass
 def GetConsumingEnumerable(self,cancellationToken=None):
  """
  GetConsumingEnumerable(self: BlockingCollection[T],cancellationToken: CancellationToken) -> IEnumerable[T]

  

   Provides a consuming System.Collections.Generics.IEnumerable for items in the collection.

  

   cancellationToken: A cancellation token to observe.

   Returns: An System.Collections.Generics.IEnumerable that removes and returns items from the collection.

  GetConsumingEnumerable(self: BlockingCollection[T]) -> IEnumerable[T]

  

   Provides a consuming System.Collections.Generics.IEnumerable for items in the collection.

   Returns: An System.Collections.Generics.IEnumerable that removes and returns items from the collection.
  """
  pass
 def Take(self,cancellationToken=None):
  """
  Take(self: BlockingCollection[T],cancellationToken: CancellationToken) -> T

  

   Takes an item from the System.Collections.Concurrent.BlockingCollection.

  

   cancellationToken: Object that can be used to cancel the take operation.

   Returns: The item removed from the collection.

  Take(self: BlockingCollection[T]) -> T

  

   Takes an item from the System.Collections.Concurrent.BlockingCollection.

   Returns: The item removed from the collection.
  """
  pass
 @staticmethod
 def TakeFromAny(collections,item,cancellationToken=None):
  """
  TakeFromAny(collections: Array[BlockingCollection[T]],cancellationToken: CancellationToken) -> (int,T)

  TakeFromAny(collections: Array[BlockingCollection[T]]) -> (int,T)
  """
  pass
 def ToArray(self):
  """
  ToArray(self: BlockingCollection[T]) -> Array[T]

  

   Copies the items from the System.Collections.Concurrent.BlockingCollection instance into a new 

    array.

  

   Returns: An array containing copies of the elements of the collection.
  """
  pass
 def TryAdd(self,item,*__args):
  """
  TryAdd(self: BlockingCollection[T],item: T,millisecondsTimeout: int) -> bool

  

   Attempts to add the specified item to the System.Collections.Concurrent.BlockingCollection 

    within the specified time period.

  

  

   item: The item to be added to the collection.

   millisecondsTimeout: The number of milliseconds to wait,or System.Threading.Timeout.Infinite (-1) to wait 

    indefinitely.

  

   Returns: true if the item could be added to the collection within the specified time; otherwise,false. 

    If the item is a duplicate,and the underlying collection does not accept duplicate items,then 

    an System.InvalidOperationException is thrown.

  

  TryAdd(self: BlockingCollection[T],item: T,millisecondsTimeout: int,cancellationToken: CancellationToken) -> bool

  

   Attempts to add the specified item to the System.Collections.Concurrent.BlockingCollection 

    within the specified time period,while observing a cancellation token.

  

  

   item: The item to be added to the collection.

   millisecondsTimeout: The number of milliseconds to wait,or System.Threading.Timeout.Infinite (-1) to wait 

    indefinitely.

  

   cancellationToken: A cancellation token to observe.

   Returns: true if the item could be added to the collection within the specified time; otherwise,false. 

    If the item is a duplicate,and the underlying collection does not accept duplicate items,then 

    an System.InvalidOperationException is thrown.

  

  TryAdd(self: BlockingCollection[T],item: T) -> bool

  

   Attempts to add the specified item to the System.Collections.Concurrent.BlockingCollection.

  

   item: The item to be added to the collection.

   Returns: true if item could be added; otherwise false. If the item is a duplicate,and the underlying 

    collection does not accept duplicate items,then an System.InvalidOperationException is thrown.

  

  TryAdd(self: BlockingCollection[T],item: T,timeout: TimeSpan) -> bool

  

   Attempts to add the specified item to the System.Collections.Concurrent.BlockingCollection.

  

   item: The item to be added to the collection.

   timeout: A System.TimeSpan that represents the number of milliseconds to wait,or a System.TimeSpan that 

    represents -1 milliseconds to wait indefinitely.

  

   Returns: true if the item could be added to the collection within the specified time span; otherwise,

    false.
  """
  pass
 @staticmethod
 def TryAddToAny(collections,item,*__args):
  """
  TryAddToAny(collections: Array[BlockingCollection[T]],item: T,millisecondsTimeout: int) -> int

  TryAddToAny(collections: Array[BlockingCollection[T]],item: T,millisecondsTimeout: int,cancellationToken: CancellationToken) -> int

  TryAddToAny(collections: Array[BlockingCollection[T]],item: T) -> int

  TryAddToAny(collections: Array[BlockingCollection[T]],item: T,timeout: TimeSpan) -> int
  """
  pass
 def TryTake(self,item,*__args):
  """
  TryTake(self: BlockingCollection[T],millisecondsTimeout: int) -> (bool,T)

  TryTake(self: BlockingCollection[T],millisecondsTimeout: int,cancellationToken: CancellationToken) -> (bool,T)

  TryTake(self: BlockingCollection[T]) -> (bool,T)

  TryTake(self: BlockingCollection[T],timeout: TimeSpan) -> (bool,T)
  """
  pass
 @staticmethod
 def TryTakeFromAny(collections,item,*__args):
  """
  TryTakeFromAny(collections: Array[BlockingCollection[T]],timeout: TimeSpan) -> (int,T)

  TryTakeFromAny(collections: Array[BlockingCollection[T]],millisecondsTimeout: int,cancellationToken: CancellationToken) -> (int,T)

  TryTakeFromAny(collections: Array[BlockingCollection[T]],millisecondsTimeout: int) -> (int,T)

  TryTakeFromAny(collections: Array[BlockingCollection[T]]) -> (int,T)
  """
  pass
 def __add__(self,*args):
  """ x.__add__(y) <==> x+yx.__add__(y) <==> x+y """
  pass
 def __contains__(self,*args):
  """ __contains__[T](enumerable: IEnumerable[T],value: T) -> bool """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 def __len__(self,*args):
  """ x.__len__() <==> len(x) """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type)

  __new__(cls: type,boundedCapacity: int)

  __new__(cls: type,collection: IProducerConsumerCollection[T],boundedCapacity: int)

  __new__(cls: type,collection: IProducerConsumerCollection[T])
  """
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 BoundedCapacity=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the bounded capacity of this System.Collections.Concurrent.BlockingCollection instance.



Get: BoundedCapacity(self: BlockingCollection[T]) -> int



"""

 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of items contained in the System.Collections.Concurrent.BlockingCollection.



Get: Count(self: BlockingCollection[T]) -> int



"""

 IsAddingCompleted=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether this System.Collections.Concurrent.BlockingCollection has been marked as complete for adding.



Get: IsAddingCompleted(self: BlockingCollection[T]) -> bool



"""

 IsCompleted=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether this System.Collections.Concurrent.BlockingCollection has been marked as complete for adding and is empty.



Get: IsCompleted(self: BlockingCollection[T]) -> bool



"""



class ConcurrentBag(object,IProducerConsumerCollection[T],IEnumerable[T],IEnumerable,ICollection,IReadOnlyCollection[T]):
 """
 ConcurrentBag[T]()

 ConcurrentBag[T](collection: IEnumerable[T])
 """
 def Add(self,item):
  """
  Add(self: ConcurrentBag[T],item: T)

   Adds an object to the System.Collections.Concurrent.ConcurrentBag.

  

   item: The object to be added to the System.Collections.Concurrent.ConcurrentBag. The value can be a 

    null reference (Nothing in Visual Basic) for reference types.
  """
  pass
 def CopyTo(self,array,index):
  """ CopyTo(self: ConcurrentBag[T],array: Array[T],index: int) """
  pass
 def GetEnumerator(self):
  """
  GetEnumerator(self: ConcurrentBag[T]) -> IEnumerator[T]

  

   Returns an enumerator that iterates through the System.Collections.Concurrent.ConcurrentBag.

   Returns: An enumerator for the contents of the System.Collections.Concurrent.ConcurrentBag.
  """
  pass
 def ToArray(self):
  """
  ToArray(self: ConcurrentBag[T]) -> Array[T]

  

   Copies the System.Collections.Concurrent.ConcurrentBag elements to a new array.

   Returns: A new array containing a snapshot of elements copied from the 

    System.Collections.Concurrent.ConcurrentBag.
  """
  pass
 def TryPeek(self,result):
  """ TryPeek(self: ConcurrentBag[T]) -> (bool,T) """
  pass
 def TryTake(self,result):
  """ TryTake(self: ConcurrentBag[T]) -> (bool,T) """
  pass
 def __add__(self,*args):
  """ x.__add__(y) <==> x+y """
  pass
 def __contains__(self,*args):
  """ __contains__[T](enumerable: IEnumerable[T],value: T) -> bool """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 def __len__(self,*args):
  """ x.__len__() <==> len(x) """
  pass
 @staticmethod
 def __new__(self,collection=None):
  """
  __new__(cls: type)

  __new__(cls: type,collection: IEnumerable[T])
  """
  pass
 def __reduce_ex__(self,*args):
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of elements contained in the System.Collections.Concurrent.ConcurrentBag.



Get: Count(self: ConcurrentBag[T]) -> int



"""

 IsEmpty=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the System.Collections.Concurrent.ConcurrentBag is empty.



Get: IsEmpty(self: ConcurrentBag[T]) -> bool



"""



class ConcurrentDictionary(object,IDictionary[TKey,TValue],ICollection[KeyValuePair[TKey,TValue]],IEnumerable[KeyValuePair[TKey,TValue]],IEnumerable,IDictionary,ICollection,IReadOnlyDictionary[TKey,TValue],IReadOnlyCollection[KeyValuePair[TKey,TValue]]):
 """
 ConcurrentDictionary[TKey,TValue]()

 ConcurrentDictionary[TKey,TValue](concurrencyLevel: int,capacity: int)

 ConcurrentDictionary[TKey,TValue](comparer: IEqualityComparer[TKey])

 ConcurrentDictionary[TKey,TValue](concurrencyLevel: int,capacity: int,comparer: IEqualityComparer[TKey])

 ConcurrentDictionary[TKey,TValue](collection: IEnumerable[KeyValuePair[TKey,TValue]])

 ConcurrentDictionary[TKey,TValue](collection: IEnumerable[KeyValuePair[TKey,TValue]],comparer: IEqualityComparer[TKey])

 ConcurrentDictionary[TKey,TValue](concurrencyLevel: int,collection: IEnumerable[KeyValuePair[TKey,TValue]],comparer: IEqualityComparer[TKey])
 """
 def AddOrUpdate(self,key,*__args):
  """
  AddOrUpdate(self: ConcurrentDictionary[TKey,TValue],key: TKey,addValue: TValue,updateValueFactory: Func[TKey,TValue,TValue]) -> TValue

  AddOrUpdate(self: ConcurrentDictionary[TKey,TValue],key: TKey,addValueFactory: Func[TKey,TValue],updateValueFactory: Func[TKey,TValue,TValue]) -> TValue
  """
  pass
 def Clear(self):
  """
  Clear(self: ConcurrentDictionary[TKey,TValue])

   Removes all keys and values from the System.Collections.Concurrent.ConcurrentDictionary.
  """
  pass
 def ContainsKey(self,key):
  """
  ContainsKey(self: ConcurrentDictionary[TKey,TValue],key: TKey) -> bool

  

   Determines whether the System.Collections.Concurrent.ConcurrentDictionary contains the specified 

    key.

  

  

   key: The key to locate in the System.Collections.Concurrent.ConcurrentDictionary.

   Returns: true if the System.Collections.Concurrent.ConcurrentDictionary contains an element with the 

    specified key; otherwise,false.
  """
  pass
 def GetEnumerator(self):
  """
  GetEnumerator(self: ConcurrentDictionary[TKey,TValue]) -> IEnumerator[KeyValuePair[TKey,TValue]]

  

   Returns an enumerator that iterates through the 

    System.Collections.Concurrent.ConcurrentDictionary.

  

   Returns: An enumerator for the System.Collections.Concurrent.ConcurrentDictionary.
  """
  pass
 def GetOrAdd(self,key,*__args):
  """
  GetOrAdd(self: ConcurrentDictionary[TKey,TValue],key: TKey,value: TValue) -> TValue

  

   Adds a key/value pair to the System.Collections.Concurrent.ConcurrentDictionary if the key does 

    not already exist.

  

  

   key: The key of the element to add.

   value: the value to be added,if the key does not already exist

   Returns: The value for the key. This will be either the existing value for the key if the key is already 

    in the dictionary,or the new value if the key was not in the dictionary.

  

  GetOrAdd(self: ConcurrentDictionary[TKey,TValue],key: TKey,valueFactory: Func[TKey,TValue]) -> TValue

  

   Adds a key/value pair to the System.Collections.Concurrent.ConcurrentDictionary if the key does 

    not already exist.

  

  

   key: The key of the element to add.

   valueFactory: The function used to generate a value for the key

   Returns: The value for the key. This will be either the existing value for the key if the key is already 

    in the dictionary,or the new value for the key as returned by valueFactory if the key was not 

    in the dictionary.
  """
  pass
 def ToArray(self):
  """
  ToArray(self: ConcurrentDictionary[TKey,TValue]) -> Array[KeyValuePair[TKey,TValue]]

  

   Copies the key and value pairs stored in the System.Collections.Concurrent.ConcurrentDictionary 

    to a new array.

  

   Returns: A new array containing a snapshot of key and value pairs copied from the 

    System.Collections.Concurrent.ConcurrentDictionary.
  """
  pass
 def TryAdd(self,key,value):
  """
  TryAdd(self: ConcurrentDictionary[TKey,TValue],key: TKey,value: TValue) -> bool

  

   Attempts to add the specified key and value to the 

    System.Collections.Concurrent.ConcurrentDictionary.

  

  

   key: The key of the element to add.

   value: The value of the element to add. The value can be a null reference (Nothing in Visual Basic) for 

    reference types.

  

   Returns: true if the key/value pair was added to the System.Collections.Concurrent.ConcurrentDictionary 

    successfully. If the key already exists,this method returns false.
  """
  pass
 def TryGetValue(self,key,value):
  """ TryGetValue(self: ConcurrentDictionary[TKey,TValue],key: TKey) -> (bool,TValue) """
  pass
 def TryRemove(self,key,value):
  """ TryRemove(self: ConcurrentDictionary[TKey,TValue],key: TKey) -> (bool,TValue) """
  pass
 def TryUpdate(self,key,newValue,comparisonValue):
  """
  TryUpdate(self: ConcurrentDictionary[TKey,TValue],key: TKey,newValue: TValue,comparisonValue: TValue) -> bool

  

   Compares the existing value for the specified key with a specified value,and if they are equal,

    updates the key with a third value.

  

  

   key: The key whose value is compared with comparisonValue and possibly replaced.

   newValue: The value that replaces the value of the element with key if the comparison results in equality.

   comparisonValue: The value that is compared to the value of the element with key.

   Returns: true if the value with key was equal to comparisonValue and replaced with newValue; otherwise,

    false.
  """
  pass
 def __contains__(self,*args):
  """
  __contains__(self: IDictionary[TKey,TValue],key: TKey) -> bool

  

   Determines whether the System.Collections.Generic.IDictionary contains an element with the 

    specified key.

  

  

   key: The key to locate in the System.Collections.Generic.IDictionary.

   Returns: true if the System.Collections.Generic.IDictionary contains an element with the key; otherwise,

    false.

  

  __contains__(self: IDictionary,key: object) -> bool

  

   Determines whether the System.Collections.IDictionary object contains an element with the 

    specified key.

  

  

   key: The key to locate in the System.Collections.IDictionary object.

   Returns: true if the System.Collections.IDictionary contains an element with the key; otherwise,false.
  """
  pass
 def __getitem__(self,*args):
  """ x.__getitem__(y) <==> x[y] """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 def __len__(self,*args):
  """ x.__len__() <==> len(x) """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type)

  __new__(cls: type,concurrencyLevel: int,capacity: int)

  __new__(cls: type,collection: IEnumerable[KeyValuePair[TKey,TValue]])

  __new__(cls: type,comparer: IEqualityComparer[TKey])

  __new__(cls: type,collection: IEnumerable[KeyValuePair[TKey,TValue]],comparer: IEqualityComparer[TKey])

  __new__(cls: type,concurrencyLevel: int,collection: IEnumerable[KeyValuePair[TKey,TValue]],comparer: IEqualityComparer[TKey])

  __new__(cls: type,concurrencyLevel: int,capacity: int,comparer: IEqualityComparer[TKey])
  """
  pass
 def __reduce_ex__(self,*args):
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 def __setitem__(self,*args):
  """ x.__setitem__(i,y) <==> x[i]= """
  pass
 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of key/value pairs contained in the System.Collections.Concurrent.ConcurrentDictionary.



Get: Count(self: ConcurrentDictionary[TKey,TValue]) -> int



"""

 IsEmpty=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the System.Collections.Concurrent.ConcurrentDictionary is empty.



Get: IsEmpty(self: ConcurrentDictionary[TKey,TValue]) -> bool



"""

 Keys=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a collection containing the keys in the System.Collections.Generic.Dictionary{TKey,TValue}.



Get: Keys(self: ConcurrentDictionary[TKey,TValue]) -> ICollection[TKey]



"""

 Values=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a collection containing the values in the System.Collections.Generic.Dictionary{TKey,TValue}.



Get: Values(self: ConcurrentDictionary[TKey,TValue]) -> ICollection[TValue]



"""



class ConcurrentQueue(object,IProducerConsumerCollection[T],IEnumerable[T],IEnumerable,ICollection,IReadOnlyCollection[T]):
 """
 ConcurrentQueue[T]()

 ConcurrentQueue[T](collection: IEnumerable[T])
 """
 def CopyTo(self,array,index):
  """ CopyTo(self: ConcurrentQueue[T],array: Array[T],index: int) """
  pass
 def Enqueue(self,item):
  """
  Enqueue(self: ConcurrentQueue[T],item: T)

   Adds an object to the end of the System.Collections.Concurrent.ConcurrentQueue.

  

   item: The object to add to the end of the System.Collections.Concurrent.ConcurrentQueue. The value can 

    be a null reference (Nothing in Visual Basic) for reference types.
  """
  pass
 def GetEnumerator(self):
  """
  GetEnumerator(self: ConcurrentQueue[T]) -> IEnumerator[T]

  

   Returns an enumerator that iterates through the System.Collections.Concurrent.ConcurrentQueue.

   Returns: An enumerator for the contents of the System.Collections.Concurrent.ConcurrentQueue.
  """
  pass
 def ToArray(self):
  """
  ToArray(self: ConcurrentQueue[T]) -> Array[T]

  

   Copies the elements stored in the System.Collections.Concurrent.ConcurrentQueue to a new array.

   Returns: A new array containing a snapshot of elements copied from the 

    System.Collections.Concurrent.ConcurrentQueue.
  """
  pass
 def TryDequeue(self,result):
  """ TryDequeue(self: ConcurrentQueue[T]) -> (bool,T) """
  pass
 def TryPeek(self,result):
  """ TryPeek(self: ConcurrentQueue[T]) -> (bool,T) """
  pass
 def __contains__(self,*args):
  """ __contains__[T](enumerable: IEnumerable[T],value: T) -> bool """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 def __len__(self,*args):
  """ x.__len__() <==> len(x) """
  pass
 @staticmethod
 def __new__(self,collection=None):
  """
  __new__(cls: type)

  __new__(cls: type,collection: IEnumerable[T])
  """
  pass
 def __reduce_ex__(self,*args):
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of elements contained in the System.Collections.Concurrent.ConcurrentQueue.



Get: Count(self: ConcurrentQueue[T]) -> int



"""

 IsEmpty=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the System.Collections.Concurrent.ConcurrentQueue is empty.



Get: IsEmpty(self: ConcurrentQueue[T]) -> bool



"""



class ConcurrentStack(object,IProducerConsumerCollection[T],IEnumerable[T],IEnumerable,ICollection,IReadOnlyCollection[T]):
 """
 ConcurrentStack[T]()

 ConcurrentStack[T](collection: IEnumerable[T])
 """
 def Clear(self):
  """
  Clear(self: ConcurrentStack[T])

   Removes all objects from the System.Collections.Concurrent.ConcurrentStack.
  """
  pass
 def CopyTo(self,array,index):
  """ CopyTo(self: ConcurrentStack[T],array: Array[T],index: int) """
  pass
 def GetEnumerator(self):
  """
  GetEnumerator(self: ConcurrentStack[T]) -> IEnumerator[T]

  

   Returns an enumerator that iterates through the System.Collections.Concurrent.ConcurrentStack.

   Returns: An enumerator for the System.Collections.Concurrent.ConcurrentStack.
  """
  pass
 def Push(self,item):
  """
  Push(self: ConcurrentStack[T],item: T)

   Inserts an object at the top of the System.Collections.Concurrent.ConcurrentStack.

  

   item: The object to push onto the System.Collections.Concurrent.ConcurrentStack. The value can be a 

    null reference (Nothing in Visual Basic) for reference types.
  """
  pass
 def PushRange(self,items,startIndex=None,count=None):
  """ PushRange(self: ConcurrentStack[T],items: Array[T],startIndex: int,count: int)PushRange(self: ConcurrentStack[T],items: Array[T]) """
  pass
 def ToArray(self):
  """
  ToArray(self: ConcurrentStack[T]) -> Array[T]

  

   Copies the items stored in the System.Collections.Concurrent.ConcurrentStack to a new array.

   Returns: A new array containing a snapshot of elements copied from the 

    System.Collections.Concurrent.ConcurrentStack.
  """
  pass
 def TryPeek(self,result):
  """ TryPeek(self: ConcurrentStack[T]) -> (bool,T) """
  pass
 def TryPop(self,result):
  """ TryPop(self: ConcurrentStack[T]) -> (bool,T) """
  pass
 def TryPopRange(self,items,startIndex=None,count=None):
  """
  TryPopRange(self: ConcurrentStack[T],items: Array[T],startIndex: int,count: int) -> int

  TryPopRange(self: ConcurrentStack[T],items: Array[T]) -> int
  """
  pass
 def __contains__(self,*args):
  """ __contains__[T](enumerable: IEnumerable[T],value: T) -> bool """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 def __len__(self,*args):
  """ x.__len__() <==> len(x) """
  pass
 @staticmethod
 def __new__(self,collection=None):
  """
  __new__(cls: type)

  __new__(cls: type,collection: IEnumerable[T])
  """
  pass
 def __reduce_ex__(self,*args):
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of elements contained in the System.Collections.Concurrent.ConcurrentStack.



Get: Count(self: ConcurrentStack[T]) -> int



"""

 IsEmpty=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the System.Collections.Concurrent.ConcurrentStack is empty.



Get: IsEmpty(self: ConcurrentStack[T]) -> bool



"""



class EnumerablePartitionerOptions(Enum,IComparable,IFormattable,IConvertible):
 """ enum (flags) EnumerablePartitionerOptions,values: NoBuffering (1),None (0) """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 NoBuffering=None
 None=None
 value__=None


class IProducerConsumerCollection(IEnumerable[T],IEnumerable,ICollection):
 # no doc
 def CopyTo(self,array,index):
  """ CopyTo(self: IProducerConsumerCollection[T],array: Array[T],index: int) """
  pass
 def ToArray(self):
  """
  ToArray(self: IProducerConsumerCollection[T]) -> Array[T]

  

   Copies the elements contained in the System.Collections.Concurrent.IProducerConsumerCollection 

    to a new array.

  

   Returns: A new array containing the elements copied from the 

    System.Collections.Concurrent.IProducerConsumerCollection.
  """
  pass
 def TryAdd(self,item):
  """
  TryAdd(self: IProducerConsumerCollection[T],item: T) -> bool

  

   Attempts to add an object to the System.Collections.Concurrent.IProducerConsumerCollection.

  

   item: The object to add to the System.Collections.Concurrent.IProducerConsumerCollection.

   Returns: true if the object was added successfully; otherwise,false.
  """
  pass
 def TryTake(self,item):
  """ TryTake(self: IProducerConsumerCollection[T]) -> (bool,T) """
  pass
 def __contains__(self,*args):
  """ __contains__[T](enumerable: IEnumerable[T],value: T) -> bool """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 def __len__(self,*args):
  """ x.__len__() <==> len(x) """
  pass

class OrderablePartitioner(Partitioner[TSource]):
 # no doc
 def GetDynamicPartitions(self):
  """
  GetDynamicPartitions(self: OrderablePartitioner[TSource]) -> IEnumerable[TSource]

  

   Creates an object that can partition the underlying collection into a variable number of 

    partitions.

  

   Returns: An object that can create partitions over the underlying data source.
  """
  pass
 def GetOrderableDynamicPartitions(self):
  """
  GetOrderableDynamicPartitions(self: OrderablePartitioner[TSource]) -> IEnumerable[KeyValuePair[Int64,TSource]]

  

   Creates an object that can partition the underlying collection into a variable number of 

    partitions.

  

   Returns: An object that can create partitions over the underlying data source.
  """
  pass
 def GetOrderablePartitions(self,partitionCount):
  """
  GetOrderablePartitions(self: OrderablePartitioner[TSource],partitionCount: int) -> IList[IEnumerator[KeyValuePair[Int64,TSource]]]

  

   Partitions the underlying collection into the specified number of orderable partitions.

  

   partitionCount: The number of partitions to create.

   Returns: A list containing partitionCount enumerators.
  """
  pass
 def GetPartitions(self,partitionCount):
  """
  GetPartitions(self: OrderablePartitioner[TSource],partitionCount: int) -> IList[IEnumerator[TSource]]

  

   Partitions the underlying collection into the given number of ordered partitions.

  

   partitionCount: The number of partitions to create.

   Returns: A list containing partitionCount enumerators.
  """
  pass
 @staticmethod
 def __new__(self,*args): #cannot find CLR constructor
  """ __new__(cls: type,keysOrderedInEachPartition: bool,keysOrderedAcrossPartitions: bool,keysNormalized: bool) """
  pass
 KeysNormalized=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether order keys are normalized.



Get: KeysNormalized(self: OrderablePartitioner[TSource]) -> bool



"""

 KeysOrderedAcrossPartitions=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether elements in an earlier partition always come before elements in a later partition.



Get: KeysOrderedAcrossPartitions(self: OrderablePartitioner[TSource]) -> bool



"""

 KeysOrderedInEachPartition=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether elements in each partition are yielded in the order of increasing keys.



Get: KeysOrderedInEachPartition(self: OrderablePartitioner[TSource]) -> bool



"""






class DataGridViewSortCompareEventArgs(HandledEventArgs):
 """
 Provides data for the System.Windows.Forms.DataGridView.SortCompare event.

 

 DataGridViewSortCompareEventArgs(dataGridViewColumn: DataGridViewColumn,cellValue1: object,cellValue2: object,rowIndex1: int,rowIndex2: int)
 """
 @staticmethod
 def __new__(self,dataGridViewColumn,cellValue1,cellValue2,rowIndex1,rowIndex2):
  """ __new__(cls: type,dataGridViewColumn: DataGridViewColumn,cellValue1: object,cellValue2: object,rowIndex1: int,rowIndex2: int) """
  pass
 CellValue1=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the value of the first cell to compare.



Get: CellValue1(self: DataGridViewSortCompareEventArgs) -> object



"""

 CellValue2=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the value of the second cell to compare.



Get: CellValue2(self: DataGridViewSortCompareEventArgs) -> object



"""

 Column=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the column being sorted.



Get: Column(self: DataGridViewSortCompareEventArgs) -> DataGridViewColumn



"""

 RowIndex1=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the index of the row containing the first cell to compare.



Get: RowIndex1(self: DataGridViewSortCompareEventArgs) -> int



"""

 RowIndex2=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the index of the row containing the second cell to compare.



Get: RowIndex2(self: DataGridViewSortCompareEventArgs) -> int



"""

 SortResult=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating the order in which the compared cells will be sorted.



Get: SortResult(self: DataGridViewSortCompareEventArgs) -> int



Set: SortResult(self: DataGridViewSortCompareEventArgs)=value

"""





# encoding: utf-8
# module System.Drawing.Text calls itself Text
# from System.Drawing,Version=4.0.0.0,Culture=neutral,PublicKeyToken=b03f5f7f11d50a3a
# by generator 1.145
""" NamespaceTracker represent a CLS namespace. """
# no imports

# no functions
# classes

class FontCollection(object,IDisposable):
 """ Provides a base class for installed and private font collections. """
 def Dispose(self):
  """
  Dispose(self: FontCollection)

   Releases all resources used by this System.Drawing.Text.FontCollection.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 Families=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the array of System.Drawing.FontFamily objects associated with this System.Drawing.Text.FontCollection.



Get: Families(self: FontCollection) -> Array[FontFamily]



"""



class GenericFontFamilies(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies a generic System.Drawing.FontFamily object.

 

 enum GenericFontFamilies,values: Monospace (2),SansSerif (1),Serif (0)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Monospace=None
 SansSerif=None
 Serif=None
 value__=None


class HotkeyPrefix(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies the type of display for hot-key prefixes that relate to text.

 

 enum HotkeyPrefix,values: Hide (2),None (0),Show (1)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Hide=None
 None=None
 Show=None
 value__=None


class InstalledFontCollection(FontCollection,IDisposable):
 """
 Represents the fonts installed on the system. This class cannot be inherited.

 

 InstalledFontCollection()
 """
 def Dispose(self):
  """
  Dispose(self: FontCollection,disposing: bool)

   Releases the unmanaged resources used by the System.Drawing.Text.FontCollection and optionally 

    releases the managed resources.

  

  

   disposing: true to release both managed and unmanaged resources; false to release only unmanaged resources.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass

class PrivateFontCollection(FontCollection,IDisposable):
 """
 Provides a collection of font families built from font files that are provided by the client application.

 

 PrivateFontCollection()
 """
 def AddFontFile(self,filename):
  """
  AddFontFile(self: PrivateFontCollection,filename: str)

   Adds a font from the specified file to this System.Drawing.Text.PrivateFontCollection.

  

   filename: A System.String that contains the file name of the font to add.
  """
  pass
 def AddMemoryFont(self,memory,length):
  """
  AddMemoryFont(self: PrivateFontCollection,memory: IntPtr,length: int)

   Adds a font contained in system memory to this System.Drawing.Text.PrivateFontCollection.

  

   memory: The memory address of the font to add.

   length: The memory length of the font to add.
  """
  pass
 def Dispose(self):
  """ Dispose(self: PrivateFontCollection,disposing: bool) """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass

class TextRenderingHint(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies the quality of text rendering.

 

 enum TextRenderingHint,values: AntiAlias (4),AntiAliasGridFit (3),ClearTypeGridFit (5),SingleBitPerPixel (2),SingleBitPerPixelGridFit (1),SystemDefault (0)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 AntiAlias=None
 AntiAliasGridFit=None
 ClearTypeGridFit=None
 SingleBitPerPixel=None
 SingleBitPerPixelGridFit=None
 SystemDefault=None
 value__=None





class Delegate(object,ICloneable,ISerializable):
 """ Represents a delegate,which is a data structure that refers to a static method or to a class instance and an instance method of that class. """
 def Call(self,*args):
  """ x.__call__(...) <==> x(...)x.__call__(...) <==> x(...) """
  pass
 def Clone(self):
  """
  Clone(self: Delegate) -> object

  

   Creates a shallow copy of the delegate.

   Returns: A shallow copy of the delegate.
  """
  pass
 @staticmethod
 def Combine(*__args):
  """
  Combine(*delegates: Array[Delegate]) -> Delegate

  

   Concatenates the invocation lists of an array of delegates.

  

   delegates: The array of delegates to combine.

   Returns: A new delegate with an invocation list that concatenates the invocation lists of the delegates 

    in the delegates array. Returns null if delegates is null,if delegates contains zero elements,

    or if every entry in delegates is null.

  

  Combine(a: Delegate,b: Delegate) -> Delegate

  

   Concatenates the invocation lists of two delegates.

  

   a: The delegate whose invocation list comes first.

   b: The delegate whose invocation list comes last.

   Returns: A new delegate with an invocation list that concatenates the invocation lists of a and b in that 

    order. Returns a if b is null,returns b if a is a null reference,and returns a null reference 

    if both a and b are null references.
  """
  pass
 def CombineImpl(self,*args):
  """
  CombineImpl(self: Delegate,d: Delegate) -> Delegate

  

   Concatenates the invocation lists of the specified multicast (combinable) delegate and the 

    current multicast (combinable) delegate.

  

  

   d: The multicast (combinable) delegate whose invocation list to append to the end of the invocation 

    list of the current multicast (combinable) delegate.

  

   Returns: A new multicast (combinable) delegate with an invocation list that concatenates the invocation 

    list of the current multicast (combinable) delegate and the invocation list of d,or the current 

    multicast (combinable) delegate if d is null.
  """
  pass
 @staticmethod
 def CreateDelegate(type,*__args):
  """
  CreateDelegate(type: Type,method: MethodInfo,throwOnBindFailure: bool) -> Delegate

  

   Creates a delegate of the specified type to represent the specified static method,with the 

    specified behavior on failure to bind.

  

  

   type: The System.Type of delegate to create.

   method: The System.Reflection.MethodInfo describing the static or instance method the delegate is to 

    represent.

  

   throwOnBindFailure: true to throw an exception if method cannot be bound; otherwise,false.

   Returns: A delegate of the specified type to represent the specified static method.

  CreateDelegate(type: Type,target: Type,method: str,ignoreCase: bool,throwOnBindFailure: bool) -> Delegate

  

   Creates a delegate of the specified type that represents the specified static method of the 

    specified class,with the specified case-sensitivity and the specified behavior on failure to 

    bind.

  

  

   type: The System.Type of delegate to create.

   target: The System.Type representing the class that implements method.

   method: The name of the static method that the delegate is to represent.

   ignoreCase: A Boolean indicating whether to ignore the case when comparing the name of the method.

   throwOnBindFailure: true to throw an exception if method cannot be bound; otherwise,false.

   Returns: A delegate of the specified type that represents the specified static method of the specified 

    class.

  

  CreateDelegate(type: Type,firstArgument: object,method: MethodInfo) -> Delegate

  

   Creates a delegate of the specified type that represents the specified static or instance 

    method,with the specified first argument.

  

  

   type: The System.Type of delegate to create.

   firstArgument: The object to which the delegate is bound,or null to treat method as static (Shared in Visual 

    Basic).

  

   method: The System.Reflection.MethodInfo describing the static or instance method the delegate is to 

    represent.

  

   Returns: A delegate of the specified type that represents the specified static or instance method.

  CreateDelegate(type: Type,method: MethodInfo) -> Delegate

  

   Creates a delegate of the specified type to represent the specified static method.

  

   type: The System.Type of delegate to create.

   method: The System.Reflection.MethodInfo describing the static or instance method the delegate is to 

    represent. Only static methods are supported in the .NET Framework version 1.0 and 1.1.

  

   Returns: A delegate of the specified type to represent the specified static method.

  CreateDelegate(type: Type,firstArgument: object,method: MethodInfo,throwOnBindFailure: bool) -> Delegate

  

   Creates a delegate of the specified type that represents the specified static or instance 

    method,with the specified first argument and the specified behavior on failure to bind.

  

  

   type: A System.Type representing the type of delegate to create.

   firstArgument: An System.Object that is the first argument of the method the delegate represents. For instance 

    methods,it must be compatible with the instance type.

  

   method: The System.Reflection.MethodInfo describing the static or instance method the delegate is to 

    represent.

  

   throwOnBindFailure: true to throw an exception if method cannot be bound; otherwise,false.

   Returns: A delegate of the specified type that represents the specified static or instance method,or 

    null if throwOnBindFailure is false and the delegate cannot be bound to method.

  

  CreateDelegate(type: Type,target: object,method: str,ignoreCase: bool) -> Delegate

  

   Creates a delegate of the specified type that represents the specified instance method to invoke 

    on the specified class instance with the specified case-sensitivity.

  

  

   type: The System.Type of delegate to create.

   target: The class instance on which method is invoked.

   method: The name of the instance method that the delegate is to represent.

   ignoreCase: A Boolean indicating whether to ignore the case when comparing the name of the method.

   Returns: A delegate of the specified type that represents the specified instance method to invoke on the 

    specified class instance.

  

  CreateDelegate(type: Type,target: object,method: str) -> Delegate

  

   Creates a delegate of the specified type that represents the specified instance method to invoke 

    on the specified class instance.

  

  

   type: The System.Type of delegate to create.

   target: The class instance on which method is invoked.

   method: The name of the instance method that the delegate is to represent.

   Returns: A delegate of the specified type that represents the specified instance method to invoke on the 

    specified class instance.

  

  CreateDelegate(type: Type,target: object,method: str,ignoreCase: bool,throwOnBindFailure: bool) -> Delegate

  

   Creates a delegate of the specified type that represents the specified instance method to invoke 

    on the specified class instance,with the specified case-sensitivity and the specified behavior 

    on failure to bind.

  

  

   type: The System.Type of delegate to create.

   target: The class instance on which method is invoked.

   method: The name of the instance method that the delegate is to represent.

   ignoreCase: A Boolean indicating whether to ignore the case when comparing the name of the method.

   throwOnBindFailure: true to throw an exception if method cannot be bound; otherwise,false.

   Returns: A delegate of the specified type that represents the specified instance method to invoke on the 

    specified class instance.

  

  CreateDelegate(type: Type,target: Type,method: str,ignoreCase: bool) -> Delegate

  

   Creates a delegate of the specified type that represents the specified static method of the 

    specified class,with the specified case-sensitivity.

  

  

   type: The System.Type of delegate to create.

   target: The System.Type representing the class that implements method.

   method: The name of the static method that the delegate is to represent.

   ignoreCase: A Boolean indicating whether to ignore the case when comparing the name of the method.

   Returns: A delegate of the specified type that represents the specified static method of the specified 

    class.

  

  CreateDelegate(type: Type,target: Type,method: str) -> Delegate

  

   Creates a delegate of the specified type that represents the specified static method of the 

    specified class.

  

  

   type: The System.Type of delegate to create.

   target: The System.Type representing the class that implements method.

   method: The name of the static method that the delegate is to represent.

   Returns: A delegate of the specified type that represents the specified static method of the specified 

    class.
  """
  pass
 def DynamicInvoke(self,args):
  """
  DynamicInvoke(self: Delegate,*args: Array[object]) -> object

  

   Dynamically invokes (late-bound) the method represented by the current delegate.

  

   args: An array of objects that are the arguments to pass to the method represented by the current 

    delegate.-or- null,if the method represented by the current delegate does not require 

    arguments.

  

   Returns: The object returned by the method represented by the delegate.
  """
  pass
 def DynamicInvokeImpl(self,*args):
  """
  DynamicInvokeImpl(self: Delegate,args: Array[object]) -> object

  

   Dynamically invokes (late-bound) the method represented by the current delegate.

  

   args: An array of objects that are the arguments to pass to the method represented by the current 

    delegate.-or- null,if the method represented by the current delegate does not require 

    arguments.

  

   Returns: The object returned by the method represented by the delegate.
  """
  pass
 def Equals(self,obj):
  """
  Equals(self: Delegate,obj: object) -> bool

  

   Determines whether the specified object and the current delegate are of the same type and share 

    the same targets,methods,and invocation list.

  

  

   obj: The object to compare with the current delegate.

   Returns: true if obj and the current delegate have the same targets,methods,and invocation list; 

    otherwise,false.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: Delegate) -> int

  

   Returns a hash code for the delegate.

   Returns: A hash code for the delegate.
  """
  pass
 def GetInvocationList(self):
  """
  GetInvocationList(self: Delegate) -> Array[Delegate]

  

   Returns the invocation list of the delegate.

   Returns: An array of delegates representing the invocation list of the current delegate.
  """
  pass
 def GetMethodImpl(self,*args):
  """
  GetMethodImpl(self: Delegate) -> MethodInfo

  

   Gets the static method represented by the current delegate.

   Returns: A System.Reflection.MethodInfo describing the static method represented by the current delegate.
  """
  pass
 def GetObjectData(self,info,context):
  """
  GetObjectData(self: Delegate,info: SerializationInfo,context: StreamingContext)

   Not supported.

  

   info: Not supported.

   context: Not supported.
  """
  pass
 def InPlaceAdd(self,*args):
  """ InPlaceAdd(self: Delegate,other: Delegate) -> Delegate """
  pass
 def InPlaceSubtract(self,*args):
  """ InPlaceSubtract(self: Delegate,other: Delegate) -> Delegate """
  pass
 @staticmethod
 def Remove(source,value):
  """
  Remove(source: Delegate,value: Delegate) -> Delegate

  

   Removes the last occurrence of the invocation list of a delegate from the invocation list of 

    another delegate.

  

  

   source: The delegate from which to remove the invocation list of value.

   value: The delegate that supplies the invocation list to remove from the invocation list of source.

   Returns: A new delegate with an invocation list formed by taking the invocation list of source and 

    removing the last occurrence of the invocation list of value,if the invocation list of value is 

    found within the invocation list of source. Returns source if value is null or if the invocation 

    list of value is not found within the invocation list of source. Returns a null reference if the 

    invocation list of value is equal to the invocation list of source or if source is a null 

    reference.
  """
  pass
 @staticmethod
 def RemoveAll(source,value):
  """
  RemoveAll(source: Delegate,value: Delegate) -> Delegate

  

   Removes all occurrences of the invocation list of a delegate from the invocation list of another 

    delegate.

  

  

   source: The delegate from which to remove the invocation list of value.

   value: The delegate that supplies the invocation list to remove from the invocation list of source.

   Returns: A new delegate with an invocation list formed by taking the invocation list of source and 

    removing all occurrences of the invocation list of value,if the invocation list of value is 

    found within the invocation list of source. Returns source if value is null or if the invocation 

    list of value is not found within the invocation list of source. Returns a null reference if the 

    invocation list of value is equal to the invocation list of source,if source contains only a 

    series of invocation lists that are equal to the invocation list of value,or if source is a 

    null reference.
  """
  pass
 def RemoveImpl(self,*args):
  """
  RemoveImpl(self: Delegate,d: Delegate) -> Delegate

  

   Removes the invocation list of a delegate from the invocation list of another delegate.

  

   d: The delegate that supplies the invocation list to remove from the invocation list of the current 

    delegate.

  

   Returns: A new delegate with an invocation list formed by taking the invocation list of the current 

    delegate and removing the invocation list of value,if the invocation list of value is found 

    within the current delegate's invocation list. Returns the current delegate if value is null or 

    if the invocation list of value is not found within the current delegate's invocation list. 

    Returns null if the invocation list of value is equal to the current delegate's invocation list.
  """
  pass
 def __call__(self,*args):
  """ x.__call__(...) <==> x(...)x.__call__(...) <==> x(...) """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __iadd__(self,*args):
  """ __iadd__(self: Delegate,other: Delegate) -> Delegate """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __isub__(self,*args):
  """ __isub__(self: Delegate,other: Delegate) -> Delegate """
  pass
 @staticmethod
 def __new__(self,*args): #cannot find CLR constructor
  """ __new__(type: type,function: object) -> object """
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 Method=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the method represented by the delegate.



Get: Method(self: Delegate) -> MethodInfo



"""

 Target=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the class instance on which the current delegate invokes the instance method.



Get: Target(self: Delegate) -> object



"""





"""Revit DB objects wrappers."""
import os.path as op

from pyrevit import HOST_APP, PyRevitException
from pyrevit.compat import safe_strtype
from pyrevit import coreutils
from pyrevit import DB
from Autodesk.Revit.DB import Element   #pylint: disable=E0401
from pyrevit.compat import get_elementid_value_func


#pylint: disable=W0703,C0302,C0103
__all__ = ('BaseWrapper', 'ElementWrapper',
           'ExternalRef', 'ProjectParameter', 'ProjectInfo',
           'XYZPoint', 'get_parameter_data_type', 'is_yesno_parameter')


class BaseWrapper(object):
    """Base revit databse object wrapper."""
    def __init__(self, obj=None):
        self._wrapped = obj

    def __repr__(self, data=None):
        pdata = {}
        if hasattr(self._wrapped, 'Id'):
            get_elementid_value = get_elementid_value_func()
            pdata['id'] = get_elementid_value(self._wrapped.Id)

        if data:
            pdata.update(data)

        datastr = ' '.join(['{0}:{1}'.format(k, v)
                            for k, v in pdata.iteritems()]) #pylint: disable=E1101
        return '<pyrevit.revit.db.{class_name} % {wrapping}{datastr}>' \
               .format(class_name=self.__class__.__name__,
                       wrapping=safe_strtype(self._wrapped),
                       datastr=(' ' + datastr) if datastr else '')

    def unwrap(self):
        return self._wrapped

    @staticmethod
    def compare_attr(src, dest, attr_name, case_sensitive=False):
        if case_sensitive:
            return safe_strtype(getattr(src, attr_name, '')).lower() == \
                   safe_strtype(getattr(dest, attr_name, '')).lower()
        else:
            return safe_strtype(getattr(src, attr_name)) == \
                   safe_strtype(getattr(dest, attr_name))

    @staticmethod
    def compare_attrs(src, dest, attr_names, case_sensitive=False):
        return [BaseWrapper.compare_attr(src,
                                         dest,
                                         x,
                                         case_sensitive=case_sensitive)
                for x in attr_names]


class ElementWrapper(BaseWrapper):
    """Revit element wrapper."""
    def __init__(self, element):
        super(ElementWrapper, self).__init__(element)
        if not isinstance(self._wrapped, DB.Element):
            raise PyRevitException('Can not wrap object that are not '
                                   'derived from Element.')

    @property
    def assoc_doc(self):
        return self._wrapped.Document

    @property
    def name(self):
        # have to use the imported Element otherwise
        # AttributeError occurs
        return Element.Name.__get__(self._wrapped)

    @property
    def symbol_name(self):
        symbol = getattr(self._wrapped, 'Symbol', None)
        if symbol:
            return Element.Name.GetValue(symbol)

    @property
    def family_name(self):
        symbol = getattr(self._wrapped, 'Symbol', None)
        if symbol:
            return Element.Name.GetValue(symbol.Family)

    @property
    def id(self):
        return self._wrapped.Id

    @property
    def unique_id(self):
        return self._wrapped.UniqueId

    @property
    def workset_id(self):
        return self._wrapped.WorksetId

    @property
    def mark(self):
        mparam = self._wrapped.Parameter[DB.BuiltInParameter.ALL_MODEL_MARK]
        return mparam.AsString() if mparam else ''

    @property
    def location(self):
        locp = getattr(self._wrapped.Location, 'Point', None)
        if locp:
            return (locp.X, locp.Y, locp.Z)
        return (None, None, None)

    @property
    def x(self):
        return self.location[0]

    @property
    def y(self):
        return self.location[1]

    @property
    def z(self):
        return self.location[2]

    def get_param(self, param_name):
        return self._wrapped.LookupParameter(param_name)

    def safe_get_param(self, param_name, default=None):
        try:
            return self._wrapped.LookupParameter(param_name)
        except Exception:
            return default


class ExternalRef(ElementWrapper):
    """External reference wraper."""
    def __init__(self, link, extref):
        super(ExternalRef, self).__init__(link)
        self._extref = extref

    @property
    def name(self):
        return DB.Element.Name.__get__(self._wrapped)

    @property
    def link(self):
        return self._wrapped

    @property
    def linktype(self):
        return self._extref.ExternalFileReferenceType

    @property
    def path(self):
        p = self._extref.GetPath()
        return DB.ModelPathUtils.ConvertModelPathToUserVisiblePath(p)

    def reload(self):
        return self._wrapped.Reload()


class ProjectParameter(BaseWrapper):
    """Project parameter wrapper."""
    def __init__(self, param_def, param_binding=None, param_ext_def=False):
        super(ProjectParameter, self).__init__()
        self.param_def = param_def
        self.param_binding = param_binding
        self.param_binding_type = self._determine_binding_type()

        self.shared = False
        self.param_ext_def = None
        self.param_guid = ''
        if param_ext_def:
            self.shared = True
            self.param_ext_def = param_ext_def
            self.param_guid = self.param_ext_def.GUID.ToString()

        self.name = self.param_def.Name

        # Revit <2017 does not have the Id parameter
        self.param_id = getattr(self.param_def, 'Id', None)

        if HOST_APP.is_exactly(2021):
            # Revit >2021 does not have the UnitType property
            self.unit_type = self.param_def.GetSpecTypeId()
            self.param_type = self.param_def.ParameterType
            self.param_group = self.param_def.ParameterGroup
        elif HOST_APP.is_newer_than(2022, or_equal=True):
            # GetSpecTypeId() Removed in Revit 2022
            self.unit_type = self.param_def.GetDataType()
            # Revit >2022 does not have the ParameterType property
            self.param_type = self.param_def.GetDataType()
            # ParameterGroup deprecated
            self.param_group = self.param_def.GetGroupTypeId().TypeId
        else:
            self.unit_type = self.param_def.UnitType
            self.param_type = self.param_def.ParameterType
            self.param_group = self.param_def.ParameterGroup


    def __eq__(self, other):
        if isinstance(self.param_def, DB.ExternalDefinition):
            return self.param_def.GUID == other.GUID or self.name == other.Name
        else:
            guid = coreutils.extract_guid(str(other))
            if guid:
                return self.param_def.GUID.ToString() == guid
            else:
                return self.name == str(other)

    def _determine_binding_type(self):
        if isinstance(self.param_binding, DB.InstanceBinding):
            return 'Instance'
        elif isinstance(self.param_binding, DB.TypeBinding):
            return 'Type'


def get_parameter_data_type(definition):
    """Get parameter data type with version compatibility.
    
    Safely retrieves parameter data type, handling both old (ParameterType) 
    and new (GetDataType) API versions.
    
    Args:
        definition: Parameter definition object (InternalDefinition, 
                   ExternalDefinition, SharedParameterDefinition, etc.)
    
    Returns:
        ForgeTypeId for Revit 2022+, ParameterType enum for older versions, 
        or None on error
    """
    try:
        if HOST_APP.is_newer_than(2022, or_equal=True):
            return definition.GetDataType()
        else:
            # Try GetDataType first (works in some 2021 versions)
            try:
                return definition.GetDataType()
            except (AttributeError, Exception):
                # Fallback to ParameterType for older versions
                return definition.ParameterType
    except Exception:
        # Final fallback
        try:
            return definition.GetDataType()
        except Exception:
            return None


def is_yesno_parameter(definition):
    """Check if parameter is Yes/No (boolean) type.
    
    Version-aware check for Yes/No parameters using appropriate API 
    (SpecTypeId.Boolean.YesNo for 2022+, ParameterType.YesNo for older).
    
    Args:
        definition: Parameter definition object
    
    Returns:
        bool: True if parameter is Yes/No type, False otherwise
    """
    try:
        data_type = get_parameter_data_type(definition)
        if data_type is None:
            return False
        
        if HOST_APP.is_newer_than(2022, or_equal=True):
            # New API: compare with SpecTypeId
            if hasattr(DB.SpecTypeId, 'Boolean') and hasattr(DB.SpecTypeId.Boolean, 'YesNo'):
                return (data_type == DB.SpecTypeId.Boolean.YesNo)
            # Fallback: check TypeId string
            if hasattr(data_type, 'TypeId'):
                return ('bool' in data_type.TypeId.lower() and 'yesno' in data_type.TypeId.lower())
        else:
            # Old API: use ParameterType enum
            if hasattr(DB.ParameterType, 'YesNo'):
                return (data_type == DB.ParameterType.YesNo)
        return False
    except Exception:
        return False


class ProjectInfo(BaseWrapper):
    """Project information."""
    def __init__(self, doc):
        super(ProjectInfo, self).__init__()
        self._doc = doc

    @property
    def name(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.Name
        else:
            return ''

    @property
    def number(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.Number
        else:
            return ''

    @property
    def address(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.Address
        else:
            return ''

    @property
    def author(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.Author
        else:
            return ''

    @property
    def building_name(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.BuildingName
        else:
            return ''

    @property
    def client_name(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.ClientName
        else:
            return ''

    @property
    def issue_date(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.IssueDate
        else:
            return ''

    @property
    def org_name(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.OrganizationName
        else:
            return ''

    @property
    def org_desc(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.OrganizationDescription
        else:
            return ''

    @property
    def status(self):
        if not self._doc.IsFamilyDocument:
            return self._doc.ProjectInformation.Status
        else:
            return ''

    @property
    def location(self):
        return op.dirname(self._doc.PathName)

    @property
    def path(self):
        return self._doc.PathName

    @property
    def filename(self):
        return op.splitext(op.basename(self.path))[0]


class XYZPoint(BaseWrapper):
    """Wrapper for XYZ point."""
    @property
    def x(self):
        return round(self._wrapped.X)

    @property
    def y(self):
        return round(self._wrapped.Y)

    @property
    def z(self):
        return round(self._wrapped.Z)

    def __repr__(self): #pylint: disable=W0222
        return super(XYZPoint, self).__repr__({'X': self.x,
                                               'Y': self.y,
                                               'Z': self.z})

    def __hash__(self):
        return hash((self.x, self.y, self.z))

    def __eq__(self, other):
        if isinstance(other, DB.XYZ):
            return self._wrapped.X == other.X \
                    and self._wrapped.Y == other.Y \
                    and self._wrapped.Z == other.Z
        elif isinstance(other, XYZPoint):
            return self.x == other.x \
                    and self.y == other.y \
                    and self.z == other.z
        return False



class InputLanguageCollection(ReadOnlyCollectionBase,ICollection,IEnumerable):
 """ Stores System.Windows.Forms.InputLanguage objects. """
 def Contains(self,value):
  """
  Contains(self: InputLanguageCollection,value: InputLanguage) -> bool

  

   Gets a value indicating whether the System.Windows.Forms.InputLanguageCollection contains the 

    specified System.Windows.Forms.InputLanguage.

  

  

   value: The System.Windows.Forms.InputLanguage to locate.

   Returns: true if the System.Windows.Forms.InputLanguage is contained in the collection; otherwise,false.
  """
  pass
 def CopyTo(self,array,index):
  """
  CopyTo(self: InputLanguageCollection,array: Array[InputLanguage],index: int)

   Copies the System.Windows.Forms.InputLanguageCollection values to a one-dimensional System.Array 

    at the specified index.

  

  

   array: The one-dimensional array that is the destination of the values copied from 

    System.Windows.Forms.InputLanguageCollection.

  

   index: The index in array where copying begins.
  """
  pass
 def IndexOf(self,value):
  """
  IndexOf(self: InputLanguageCollection,value: InputLanguage) -> int

  

   Returns the index of an System.Windows.Forms.InputLanguage in the 

    System.Windows.Forms.InputLanguageCollection.

  

  

   value: The System.Windows.Forms.InputLanguage to locate.

   Returns: The index of the System.Windows.Forms.InputLanguage in the 

    System.Windows.Forms.InputLanguageCollection,if found; otherwise,-1.
  """
  pass
 def __getitem__(self,*args):
  """ x.__getitem__(y) <==> x[y] """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 InnerList=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the list of elements contained in the System.Collections.ReadOnlyCollectionBase instance.



"""





class IRevertibleChangeTracking(IChangeTracking):
 """ Provides support for rolling back the changes """
 def RejectChanges(self):
  """
  RejectChanges(self: IRevertibleChangeTracking)

    """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



from typing import Tuple, Set, Iterable, List


class Field:
    @property
    def FieldName(self) -> str: ...
    @property
    def Schema(self) -> Schema: ...
    @property
    def SubSchemaGUID(self) -> Guid: ...
    @property
    def ContainerType(self) -> ContainerType: ...
    @property
    def Documentation(self) -> str: ...
    @property
    def ValueType(self) -> Type: ...
    @property
    def KeyType(self) -> Type: ...
    @property
    def SubSchema(self) -> Schema: ...
    @property
    def UnitType(self) -> UnitType: ...
    @property
    def IsValidObject(self) -> bool: ...
    def CompatibleDisplayUnitType(self, type: DisplayUnitType) -> bool: ...
    def CompatibleUnit(self, unitTypeId: ForgeTypeId) -> bool: ...
    def SubEntityReadAccessGranted(self) -> bool: ...
    def SubEntityWriteAccessGranted(self) -> bool: ...
    def GetSpecTypeId(self) -> ForgeTypeId: ...
    def Dispose(self) -> None: ...


class Entity:
    @overload
    def __init__(self, schemaGUID: Guid): ...
    @overload
    def __init__(self, schema: Schema): ...
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, other: Entity): ...
    @overload
    def Get(self, fieldName: str, unitTypeId: ForgeTypeId) -> FieldType: ...
    @overload
    def Get(self, fieldName: str, displayUnits: DisplayUnitType) -> FieldType: ...
    @overload
    def Get(self, field: Field, unitTypeId: ForgeTypeId) -> FieldType: ...
    @overload
    def Get(self, field: Field, displayUnits: DisplayUnitType) -> FieldType: ...
    @overload
    def Get(self, fieldName: str) -> FieldType: ...
    @overload
    def Get(self, field: Field) -> FieldType: ...
    @overload
    def Set(self, fieldName: str, value: FieldType, unitTypeId: ForgeTypeId) -> None: ...
    @overload
    def Set(self, fieldName: str, value: FieldType, displayUnits: DisplayUnitType) -> None: ...
    @overload
    def Set(self, field: Field, value: FieldType, unitTypeId: ForgeTypeId) -> None: ...
    @overload
    def Set(self, field: Field, value: FieldType, displayUnits: DisplayUnitType) -> None: ...
    @overload
    def Set(self, fieldName: str, value: FieldType) -> None: ...
    @overload
    def Set(self, field: Field, value: FieldType) -> None: ...
    @property
    def Schema(self) -> Schema: ...
    @property
    def SchemaGUID(self) -> Guid: ...
    @property
    def IsValidObject(self) -> bool: ...
    def IsValid(self) -> bool: ...
    def ReadAccessGranted(self) -> bool: ...
    def WriteAccessGranted(self) -> bool: ...
    def RecognizedField(self, field: Field) -> bool: ...
    @overload
    def Clear(self, fieldName: str) -> None: ...
    @overload
    def Clear(self, field: Field) -> None: ...
    def Dispose(self) -> None: ...


class AccessLevel:
    Public = 1
    Vendor = 2
    Application = 3


class ContainerType:
    Simple = 0
    Array = 1
    Map = 2


class DataStorage(Element):
    def Create(doc: Document) -> DataStorage: ...


class Schema:
    @property
    def GUID(self) -> Guid: ...
    @property
    def SchemaName(self) -> str: ...
    @property
    def ReadAccessLevel(self) -> AccessLevel: ...
    @property
    def WriteAccessLevel(self) -> AccessLevel: ...
    @property
    def ApplicationGUID(self) -> Guid: ...
    @property
    def VendorId(self) -> str: ...
    @property
    def Documentation(self) -> str: ...
    @property
    def IsValidObject(self) -> bool: ...
    def Lookup(guid: Guid) -> Schema: ...
    def ListSchemas() -> List[Schema]: ...
    def EraseSchemaAndAllEntities(schema: Schema, overrideWriteAccessWithUserPermission: bool) -> None: ...
    def GetField(self, name: str) -> Field: ...
    def ListFields(self) -> List[Field]: ...
    def ReadAccessGranted(self) -> bool: ...
    def WriteAccessGranted(self) -> bool: ...
    def Dispose(self) -> None: ...


class FieldBuilder:
    @overload
    def __init__(self, field: Field, builder: SchemaBuilder): ...
    @overload
    def __init__(self, other: FieldBuilder): ...
    @property
    def IsValidObject(self) -> bool: ...
    def SetUnitType(self, units: UnitType) -> FieldBuilder: ...
    def SetSpec(self, specTypeId: ForgeTypeId) -> FieldBuilder: ...
    def SetSubSchemaGUID(self, guid: Guid) -> FieldBuilder: ...
    def SetDocumentation(self, documentation: str) -> FieldBuilder: ...
    def Ready(self) -> bool: ...
    def NeedsUnits(self) -> bool: ...
    def NeedsSubSchemaGUID(self) -> bool: ...
    def Dispose(self) -> None: ...


class SchemaBuilder:
    def __init__(self, guid: Guid): ...
    @property
    def IsValidObject(self) -> bool: ...
    def GUIDIsValid(guid: Guid) -> bool: ...
    def VendorIdIsValid(vendorId: str) -> bool: ...
    def AddSimpleField(self, fieldName: str, fieldType: Type) -> FieldBuilder: ...
    def AddArrayField(self, fieldName: str, fieldType: Type) -> FieldBuilder: ...
    def AddMapField(self, fieldName: str, keyType: Type, valueType: Type) -> FieldBuilder: ...
    def SetReadAccessLevel(self, readAccessLevel: AccessLevel) -> SchemaBuilder: ...
    def SetWriteAccessLevel(self, writeAccessLevel: AccessLevel) -> SchemaBuilder: ...
    def SetSchemaName(self, schemaName: str) -> SchemaBuilder: ...
    def SetDocumentation(self, documentation: str) -> SchemaBuilder: ...
    def SetApplicationGUID(self, applicationGUID: Guid) -> SchemaBuilder: ...
    def SetVendorId(self, vendorId: str) -> SchemaBuilder: ...
    def Finish(self) -> Schema: ...
    def Ready(self) -> bool: ...
    def AcceptableName(self, name: str) -> bool: ...
    def Dispose(self) -> None: ...


class ExtensibleStorageFilter(ElementQuickFilter):
    def __init__(self, schemaGuid: Guid): ...
    @property
    def SchemaGuid(self) -> Guid: ...



class IDropTarget:
 """ Defines mouse events. """
 def OnDragDrop(self,e):
  """
  OnDragDrop(self: IDropTarget,e: DragEventArgs)

   Raises the System.Windows.Forms.Control.DragDrop event.

  

   e: A System.Windows.Forms.DragEventArgs that contains the event data.
  """
  pass
 def OnDragEnter(self,e):
  """
  OnDragEnter(self: IDropTarget,e: DragEventArgs)

   Raises the System.Windows.Forms.Control.DragEnter event.

  

   e: A System.Windows.Forms.DragEventArgs that contains the event data.
  """
  pass
 def OnDragLeave(self,e):
  """
  OnDragLeave(self: IDropTarget,e: EventArgs)

   Raises the System.Windows.Forms.Control.DragLeave event.

  

   e: An System.EventArgs that contains the event data.
  """
  pass
 def OnDragOver(self,e):
  """
  OnDragOver(self: IDropTarget,e: DragEventArgs)

   Raises the System.Windows.Forms.Control.DragOver event.

  

   e: A System.Windows.Forms.DragEventArgs that contains the event data.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class TreeViewEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.TreeView.AfterCheck,System.Windows.Forms.TreeView.AfterCollapse,System.Windows.Forms.TreeView.AfterExpand,or System.Windows.Forms.TreeView.AfterSelect events of a System.Windows.Forms.TreeView control.

 

 TreeViewEventArgs(node: TreeNode)

 TreeViewEventArgs(node: TreeNode,action: TreeViewAction)
 """
 @staticmethod
 def __new__(self,node,action=None):
  """
  __new__(cls: type,node: TreeNode)

  __new__(cls: type,node: TreeNode,action: TreeViewAction)
  """
  pass
 Action=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the type of action that raised the event.



Get: Action(self: TreeViewEventArgs) -> TreeViewAction



"""

 Node=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the tree node that has been checked,expanded,collapsed,or selected.



Get: Node(self: TreeViewEventArgs) -> TreeNode



"""





class ResolveEventArgs(EventArgs):
 """
 Provides data for loader resolution events,such as the System.AppDomain.TypeResolve,System.AppDomain.ResourceResolve,System.AppDomain.ReflectionOnlyAssemblyResolve,and System.AppDomain.AssemblyResolve events.

 

 ResolveEventArgs(name: str)

 ResolveEventArgs(name: str,requestingAssembly: Assembly)
 """
 @staticmethod
 def __new__(self,name,requestingAssembly=None):
  """
  __new__(cls: type,name: str)

  __new__(cls: type,name: str,requestingAssembly: Assembly)
  """
  pass
 Name=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the name of the item to resolve.



Get: Name(self: ResolveEventArgs) -> str



"""

 RequestingAssembly=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the assembly whose dependency is being resolved.



Get: RequestingAssembly(self: ResolveEventArgs) -> Assembly



"""





class ToolStripProfessionalRenderer(ToolStripRenderer):
 """
 Handles the painting functionality for System.Windows.Forms.ToolStrip objects,applying a custom palette and a streamlined style.

 

 ToolStripProfessionalRenderer()

 ToolStripProfessionalRenderer(professionalColorTable: ProfessionalColorTable)
 """
 @staticmethod
 def __new__(self,professionalColorTable=None):
  """
  __new__(cls: type)

  __new__(cls: type,professionalColorTable: ProfessionalColorTable)
  """
  pass
 ColorTable=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the color palette used for painting.



Get: ColorTable(self: ToolStripProfessionalRenderer) -> ProfessionalColorTable



"""

 RoundedEdges=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether edges of controls have a rounded rather than a square or sharp appearance.



Get: RoundedEdges(self: ToolStripProfessionalRenderer) -> bool



Set: RoundedEdges(self: ToolStripProfessionalRenderer)=value

"""





class TreeViewHitTestInfo(object):
 """
 Contains information about an area of a System.Windows.Forms.TreeView control or a System.Windows.Forms.TreeNode.

 

 TreeViewHitTestInfo(hitNode: TreeNode,hitLocation: TreeViewHitTestLocations)
 """
 @staticmethod
 def __new__(self,hitNode,hitLocation):
  """ __new__(cls: type,hitNode: TreeNode,hitLocation: TreeViewHitTestLocations) """
  pass
 Location=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the location of a hit test on a System.Windows.Forms.TreeView control,in relation to the System.Windows.Forms.TreeView and the nodes it contains.



Get: Location(self: TreeViewHitTestInfo) -> TreeViewHitTestLocations



"""

 Node=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Windows.Forms.TreeNode at the position indicated by a hit test of a System.Windows.Forms.TreeView control.



Get: Node(self: TreeViewHitTestInfo) -> TreeNode



"""





class DefaultCharSetAttribute(Attribute,_Attribute):
 """
 Specifies the value of the System.Runtime.InteropServices.CharSet enumeration. This class cannot be inherited.

 

 DefaultCharSetAttribute(charSet: CharSet)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,charSet):
  """ __new__(cls: type,charSet: CharSet) """
  pass
 CharSet=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the default value of System.Runtime.InteropServices.CharSet for any call to System.Runtime.InteropServices.DllImportAttribute.



Get: CharSet(self: DefaultCharSetAttribute) -> CharSet



"""





class KeyPressEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.Control.KeyPress event.

 

 KeyPressEventArgs(keyChar: Char)
 """
 @staticmethod
 def __new__(self,keyChar):
  """ __new__(cls: type,keyChar: Char) """
  pass
 Handled=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether the System.Windows.Forms.Control.KeyPress event was handled.



Get: Handled(self: KeyPressEventArgs) -> bool



Set: Handled(self: KeyPressEventArgs)=value

"""

 KeyChar=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the character corresponding to the key pressed.



Get: KeyChar(self: KeyPressEventArgs) -> Char



Set: KeyChar(self: KeyPressEventArgs)=value

"""





class DataGridViewRowPostPaintEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.DataGridView.RowPostPaint event.

 

 DataGridViewRowPostPaintEventArgs(dataGridView: DataGridView,graphics: Graphics,clipBounds: Rectangle,rowBounds: Rectangle,rowIndex: int,rowState: DataGridViewElementStates,errorText: str,inheritedRowStyle: DataGridViewCellStyle,isFirstDisplayedRow: bool,isLastVisibleRow: bool)
 """
 def DrawFocus(self,bounds,cellsPaintSelectionBackground):
  """
  DrawFocus(self: DataGridViewRowPostPaintEventArgs,bounds: Rectangle,cellsPaintSelectionBackground: bool)

   Draws the focus rectangle around the specified bounds.

  

   bounds: A System.Drawing.Rectangle that specifies the focus area.

   cellsPaintSelectionBackground: true to use the System.Windows.Forms.DataGridViewCellStyle.SelectionBackColor property of the 

    System.Windows.Forms.DataGridViewRow.InheritedStyle property to determine the color of the focus 

    rectangle; false to use the System.Windows.Forms.DataGridViewCellStyle.BackColor property of the 

    System.Windows.Forms.DataGridViewRow.InheritedStyle property.
  """
  pass
 def PaintCells(self,clipBounds,paintParts):
  """
  PaintCells(self: DataGridViewRowPostPaintEventArgs,clipBounds: Rectangle,paintParts: DataGridViewPaintParts)

   Paints the specified cell parts for the area in the specified bounds.

  

   clipBounds: A System.Drawing.Rectangle that specifies the area of the System.Windows.Forms.DataGridView to 

    be painted.

  

   paintParts: A bitwise combination of System.Windows.Forms.DataGridViewPaintParts values specifying the parts 

    to paint.
  """
  pass
 def PaintCellsBackground(self,clipBounds,cellsPaintSelectionBackground):
  """
  PaintCellsBackground(self: DataGridViewRowPostPaintEventArgs,clipBounds: Rectangle,cellsPaintSelectionBackground: bool)

   Paints the cell backgrounds for the area in the specified bounds.

  

   clipBounds: A System.Drawing.Rectangle that specifies the area of the System.Windows.Forms.DataGridView to 

    be painted.

  

   cellsPaintSelectionBackground: true to paint the background of the specified bounds with the color of the 

    System.Windows.Forms.DataGridViewCellStyle.SelectionBackColor property of the 

    System.Windows.Forms.DataGridViewRow.InheritedStyle; false to paint the background of the 

    specified bounds with the color of the System.Windows.Forms.DataGridViewCellStyle.BackColor 

    property of the System.Windows.Forms.DataGridViewRow.InheritedStyle.
  """
  pass
 def PaintCellsContent(self,clipBounds):
  """
  PaintCellsContent(self: DataGridViewRowPostPaintEventArgs,clipBounds: Rectangle)

   Paints the cell contents for the area in the specified bounds.

  

   clipBounds: A System.Drawing.Rectangle that specifies the area of the System.Windows.Forms.DataGridView to 

    be painted.
  """
  pass
 def PaintHeader(self,*__args):
  """
  PaintHeader(self: DataGridViewRowPostPaintEventArgs,paintParts: DataGridViewPaintParts)

   Paints the specified parts of the row header of the current row.

  

   paintParts: A bitwise combination of System.Windows.Forms.DataGridViewPaintParts values specifying the parts 

    to paint.

  

  PaintHeader(self: DataGridViewRowPostPaintEventArgs,paintSelectionBackground: bool)

   Paints the entire row header of the current System.Windows.Forms.DataGridViewRow.

  

   paintSelectionBackground: true to paint the row header with the color of the 

    System.Windows.Forms.DataGridViewCellStyle.SelectionBackColor property of the 

    System.Windows.Forms.DataGridViewRow.InheritedStyle; false to paint the row header with the 

    System.Windows.Forms.DataGridViewCellStyle.BackColor of the 

    System.Windows.Forms.DataGridView.RowHeadersDefaultCellStyle property.
  """
  pass
 @staticmethod
 def __new__(self,dataGridView,graphics,clipBounds,rowBounds,rowIndex,rowState,errorText,inheritedRowStyle,isFirstDisplayedRow,isLastVisibleRow):
  """ __new__(cls: type,dataGridView: DataGridView,graphics: Graphics,clipBounds: Rectangle,rowBounds: Rectangle,rowIndex: int,rowState: DataGridViewElementStates,errorText: str,inheritedRowStyle: DataGridViewCellStyle,isFirstDisplayedRow: bool,isLastVisibleRow: bool) """
  pass
 ClipBounds=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the area of the System.Windows.Forms.DataGridView that needs to be repainted.



Get: ClipBounds(self: DataGridViewRowPostPaintEventArgs) -> Rectangle



Set: ClipBounds(self: DataGridViewRowPostPaintEventArgs)=value

"""

 ErrorText=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a string that represents an error message for the current System.Windows.Forms.DataGridViewRow.



Get: ErrorText(self: DataGridViewRowPostPaintEventArgs) -> str



"""

 Graphics=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Drawing.Graphics used to paint the current System.Windows.Forms.DataGridViewRow.



Get: Graphics(self: DataGridViewRowPostPaintEventArgs) -> Graphics



"""

 InheritedRowStyle=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the cell style applied to the current System.Windows.Forms.DataGridViewRow.



Get: InheritedRowStyle(self: DataGridViewRowPostPaintEventArgs) -> DataGridViewCellStyle



"""

 IsFirstDisplayedRow=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the current row is the first row displayed in the System.Windows.Forms.DataGridView.



Get: IsFirstDisplayedRow(self: DataGridViewRowPostPaintEventArgs) -> bool



"""

 IsLastVisibleRow=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the current row is the last visible row displayed in the System.Windows.Forms.DataGridView.



Get: IsLastVisibleRow(self: DataGridViewRowPostPaintEventArgs) -> bool



"""

 RowBounds=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get the bounds of the current System.Windows.Forms.DataGridViewRow.



Get: RowBounds(self: DataGridViewRowPostPaintEventArgs) -> Rectangle



"""

 RowIndex=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the index of the current System.Windows.Forms.DataGridViewRow.



Get: RowIndex(self: DataGridViewRowPostPaintEventArgs) -> int



"""

 State=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the state of the current System.Windows.Forms.DataGridViewRow.



Get: State(self: DataGridViewRowPostPaintEventArgs) -> DataGridViewElementStates



"""





class ConvertEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.Binding.Format and System.Windows.Forms.Binding.Parse events.

 

 ConvertEventArgs(value: object,desiredType: Type)
 """
 @staticmethod
 def __new__(self,value,desiredType):
  """ __new__(cls: type,value: object,desiredType: Type) """
  pass
 DesiredType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the data type of the desired value.



Get: DesiredType(self: ConvertEventArgs) -> Type



"""

 Value=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the value of the System.Windows.Forms.ConvertEventArgs.



Get: Value(self: ConvertEventArgs) -> object



Set: Value(self: ConvertEventArgs)=value

"""





class DllImportSearchPath(Enum,IComparable,IFormattable,IConvertible):
 """ enum (flags) DllImportSearchPath,values: ApplicationDirectory (512),AssemblyDirectory (2),LegacyBehavior (0),SafeDirectories (4096),System32 (2048),UseDllDirectoryForDependencies (256),UserDirectories (1024) """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 ApplicationDirectory=None
 AssemblyDirectory=None
 LegacyBehavior=None
 SafeDirectories=None
 System32=None
 UseDllDirectoryForDependencies=None
 UserDirectories=None
 value__=None




from typing import Tuple, Set, Iterable, List


class IExternalServer:
    def GetServerId(self) -> Guid: ...
    def GetServiceId(self) -> ExternalServiceId: ...
    def GetName(self) -> str: ...
    def GetVendorId(self) -> str: ...
    def GetDescription(self) -> str: ...


class IExternalData:


class ExternalServiceOptions:
    def __init__(self): ...
    @property
    def IsPublic(self) -> bool: ...
    @IsPublic.setter
    def IsPublic(self, isPublic: bool) -> None: ...
    @property
    def IsRecordable(self) -> bool: ...
    @IsRecordable.setter
    def IsRecordable(self, isRecordable: bool) -> None: ...
    @property
    def IsSelfsynchronizing(self) -> bool: ...
    @IsSelfsynchronizing.setter
    def IsSelfsynchronizing(self, isSelfsynchronizing: bool) -> None: ...
    @property
    def IsValidObject(self) -> bool: ...
    def Dispose(self) -> None: ...


class ExternalService:
    @property
    def ServiceId(self) -> ExternalServiceId: ...
    @property
    def Name(self) -> str: ...
    @property
    def Description(self) -> str: ...
    @property
    def VendorId(self) -> str: ...
    @property
    def IsSerializable(self) -> bool: ...
    @property
    def NumberOfServers(self) -> int: ...
    @property
    def IsValidObject(self) -> bool: ...
    def GetOptions(self) -> ExternalServiceOptions: ...
    def GetPublicAccessKey(self) -> Guid: ...
    def AddServer(self, server: IExternalServer) -> None: ...
    def RemoveServer(self, serverId: Guid) -> None: ...
    def GetServer(self, serverId: Guid) -> IExternalServer: ...
    def GetDefaultServerId(self) -> Guid: ...
    def GetRegisteredServerIds(self) -> List[Guid]: ...
    def IsRegisteredServerId(self, serverId: Guid) -> bool: ...
    def Dispose(self) -> None: ...


class ExternalServiceId(GuidEnum):
    def __init__(self, guid: Guid): ...
    def op_Equality(first: ExternalServiceId, second: ExternalServiceId) -> bool: ...
    def op_Inequality(first: ExternalServiceId, second: ExternalServiceId) -> bool: ...


class ExternalServices:


class BuiltInExternalServices:
    @property
    def ATFTranslationService() -> ExternalServiceId: ...
    @property
    def AlignmentService() -> ExternalServiceId: ...
    @property
    def CloudExternalService() -> ExternalServiceId: ...
    @property
    def CodeCheckingParameterService() -> ExternalServiceId: ...
    @property
    def ConnectionTypeChangedService() -> ExternalServiceId: ...
    @property
    def DirectContext3DService() -> ExternalServiceId: ...
    @property
    def DuctFittingAndAccessoryPressureDropService() -> ExternalServiceId: ...
    @property
    def DuctFittingAndAccessoryPressureDropUIService() -> ExternalServiceId: ...
    @property
    def DuctPressureDropService() -> ExternalServiceId: ...
    @property
    def EntitlementExternalService() -> ExternalServiceId: ...
    @property
    def ExternalResourceService() -> ExternalServiceId: ...
    @property
    def ExternalResourceUIService() -> ExternalServiceId: ...
    @property
    def FramingProfileService() -> ExternalServiceId: ...
    @property
    def HoleDefinitionService() -> ExternalServiceId: ...
    @property
    def IFCExporterService() -> ExternalServiceId: ...
    @property
    def IFCImporterService() -> ExternalServiceId: ...
    @property
    def MemberForcesService() -> ExternalServiceId: ...
    @property
    def ModifyConnectionParametersService() -> ExternalServiceId: ...
    @property
    def NavisworksExporterService() -> ExternalServiceId: ...
    @property
    def PathOfTravelCalculationService() -> ExternalServiceId: ...
    @property
    def PipeFittingAndAccessoryPressureDropService() -> ExternalServiceId: ...
    @property
    def PipeFittingAndAccessoryPressureDropUIService() -> ExternalServiceId: ...
    @property
    def PipeFrictionFactorService() -> ExternalServiceId: ...
    @property
    def PipePlumbingFixtureFlowService() -> ExternalServiceId: ...
    @property
    def PipePressureDropService() -> ExternalServiceId: ...
    @property
    def RebarUpdateService() -> ExternalServiceId: ...
    @property
    def SiteInsertService() -> ExternalServiceId: ...
    @property
    def SiteLinkerUIService() -> ExternalServiceId: ...
    @property
    def SnappingService() -> ExternalServiceId: ...
    @property
    def StructuralSectionsService() -> ExternalServiceId: ...


class ExternalServiceRegistry:
    @overload
    def RegisterService(service: ISingleServerService, defaultServerId: Guid, options: ExternalServiceOptions) -> Guid: ...
    @overload
    def RegisterService(service: IMultiServerService, options: ExternalServiceOptions, policy: ExecutionPolicy) -> Guid: ...
    @overload
    def RegisterService(service: ISingleServerService, options: ExternalServiceOptions) -> Guid: ...
    def GetService(serviceId: ExternalServiceId) -> ExternalService: ...
    def GetServices() -> List[ExternalService]: ...
    @overload
    def ExecuteService(executionKey: Guid, serverId: Guid, data: IExternalData) -> ExternalServiceResult: ...
    @overload
    def ExecuteService(executionKey: Guid, document: Document, data: IExternalData) -> ExternalServiceResult: ...
    @overload
    def ExecuteService(executionKey: Guid, data: IExternalData) -> ExternalServiceResult: ...


class IExternalService:
    def GetServiceId(self) -> ExternalServiceId: ...
    def GetName(self) -> str: ...
    def GetVendorId(self) -> str: ...
    def GetDescription(self) -> str: ...
    def IsValidServer(self, server: IExternalServer) -> bool: ...
    def Execute(self, server: IExternalServer, document: Document, data: IExternalData) -> bool: ...
    def OnServersDisparity(self, document: Document, oldServers: List[Guid]) -> DisparityResponse: ...
    def OnServersChanged(self, document: Document, cause: ServerChangeCause, oldServers: List[Guid]) -> None: ...


class IMultiServerService:
    def CanExecute(self, server: IExternalServer, document: Document, data: IExternalData) -> bool: ...


class ISingleServerService:


class ServerChangeCause:
    ImposedChange = 0
    UserChange = 1


class ExternalServiceResult:
    Succeeded = 0
    Failed = 1
    Unhandled = 2


class ExecutionPolicy:
    FirstApplicableServer = 0
    AllApplicableServers = 1


class DisparityResponse:
    DoNothing = 0
    ApplyDefaults = 1
    LetUserDecide = 2


class MultiServerService(ExternalService):
    @property
    def ExecutionPolicy(self) -> ExecutionPolicy: ...
    @overload
    def GetActiveServerIds(self, document: Document) -> List[Guid]: ...
    @overload
    def GetActiveServerIds(self) -> List[Guid]: ...
    @overload
    def SetActiveServers(self, serverIds: List[Guid], document: Document) -> None: ...
    @overload
    def SetActiveServers(self, serverIds: List[Guid]) -> None: ...


class SingleServerService(ExternalService):
    @overload
    def GetActiveServerId(self, document: Document) -> Guid: ...
    @overload
    def GetActiveServerId(self) -> Guid: ...
    @overload
    def SetActiveServer(self, serverId: Guid, document: Document) -> None: ...
    @overload
    def SetActiveServer(self, serverId: Guid) -> None: ...
    def UnsetActiveServer(self, document: Document) -> None: ...



class INotifyDataErrorInfo:
 # no doc
 def GetErrors(self,propertyName):
  """ GetErrors(self: INotifyDataErrorInfo,propertyName: str) -> IEnumerable """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 HasErrors=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: HasErrors(self: INotifyDataErrorInfo) -> bool



"""


 ErrorsChanged=None




class ITypedList:
 """ Provides functionality to discover the schema for a bindable list,where the properties available for binding differ from the public properties of the object to bind to. """
 def GetItemProperties(self,listAccessors):
  """
  GetItemProperties(self: ITypedList,listAccessors: Array[PropertyDescriptor]) -> PropertyDescriptorCollection

  

   Returns the System.ComponentModel.PropertyDescriptorCollection that represents the properties on 

    each item used to bind data.

  

  

   listAccessors: An array of System.ComponentModel.PropertyDescriptor objects to find in the collection as 

    bindable. This can be null.

  

   Returns: The System.ComponentModel.PropertyDescriptorCollection that represents the properties on each 

    item used to bind data.
  """
  pass
 def GetListName(self,listAccessors):
  """
  GetListName(self: ITypedList,listAccessors: Array[PropertyDescriptor]) -> str

  

   Returns the name of the list.

  

   listAccessors: An array of System.ComponentModel.PropertyDescriptor objects,for which the list name is 

    returned. This can be null.

  

   Returns: The name of the list.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class HttpRequestHeader(Enum,IComparable,IFormattable,IConvertible):
 """
 The HTTP headers that may be specified in a client request.

 

 enum HttpRequestHeader,values: Accept (20),AcceptCharset (21),AcceptEncoding (22),AcceptLanguage (23),Allow (10),Authorization (24),CacheControl (0),Connection (1),ContentEncoding (13),ContentLanguage (14),ContentLength (11),ContentLocation (15),ContentMd5 (16),ContentRange (17),ContentType (12),Cookie (25),Date (2),Expect (26),Expires (18),From (27),Host (28),IfMatch (29),IfModifiedSince (30),IfNoneMatch (31),IfRange (32),IfUnmodifiedSince (33),KeepAlive (3),LastModified (19),MaxForwards (34),Pragma (4),ProxyAuthorization (35),Range (37),Referer (36),Te (38),Trailer (5),TransferEncoding (6),Translate (39),Upgrade (7),UserAgent (40),Via (8),Warning (9)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Accept=None
 AcceptCharset=None
 AcceptEncoding=None
 AcceptLanguage=None
 Allow=None
 Authorization=None
 CacheControl=None
 Connection=None
 ContentEncoding=None
 ContentLanguage=None
 ContentLength=None
 ContentLocation=None
 ContentMd5=None
 ContentRange=None
 ContentType=None
 Cookie=None
 Date=None
 Expect=None
 Expires=None
 From=None
 Host=None
 IfMatch=None
 IfModifiedSince=None
 IfNoneMatch=None
 IfRange=None
 IfUnmodifiedSince=None
 KeepAlive=None
 LastModified=None
 MaxForwards=None
 Pragma=None
 ProxyAuthorization=None
 Range=None
 Referer=None
 Te=None
 Trailer=None
 TransferEncoding=None
 Translate=None
 Upgrade=None
 UserAgent=None
 value__=None
 Via=None
 Warning=None




class SafeHandle(CriticalFinalizerObject,IDisposable):
 """ Represents a wrapper class for operating system handles. This class must be inherited. """
 def Close(self):
  """
  Close(self: SafeHandle)

   Marks the handle for releasing and freeing resources.
  """
  pass
 def DangerousAddRef(self,success):
  """
  DangerousAddRef(self: SafeHandle,success: bool) -> bool

  

   Manually increments the reference counter on System.Runtime.InteropServices.SafeHandle instances.

  

   success: true if the reference counter was successfully incremented; otherwise,false.
  """
  pass
 def DangerousGetHandle(self):
  """
  DangerousGetHandle(self: SafeHandle) -> IntPtr

  

   Returns the value of the System.Runtime.InteropServices.SafeHandle.handle field.

   Returns: An IntPtr representing the value of the System.Runtime.InteropServices.SafeHandle.handle field. 

    If the handle has been marked invalid with 

    System.Runtime.InteropServices.SafeHandle.SetHandleAsInvalid,this method still returns the 

    original handle value,which can be a stale value.
  """
  pass
 def DangerousRelease(self):
  """
  DangerousRelease(self: SafeHandle)

   Manually decrements the reference counter on a System.Runtime.InteropServices.SafeHandle 

    instance.
  """
  pass
 def Dispose(self):
  """
  Dispose(self: SafeHandle)

   Releases all resources used by the System.Runtime.InteropServices.SafeHandle class.
  """
  pass
 def ReleaseHandle(self,*args):
  """
  ReleaseHandle(self: SafeHandle) -> bool

  

   When overridden in a derived class,executes the code required to free the handle.

   Returns: true if the handle is released successfully; otherwise,in the event of a catastrophic failure,

    false. In this case,it generates a releaseHandleFailed MDA Managed Debugging Assistant.
  """
  pass
 def SetHandle(self,*args):
  """
  SetHandle(self: SafeHandle,handle: IntPtr)

   Sets the handle to the specified pre-existing handle.

  

   handle: The pre-existing handle to use.
  """
  pass
 def SetHandleAsInvalid(self):
  """
  SetHandleAsInvalid(self: SafeHandle)

   Marks a handle as no longer used.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,*args): #cannot find CLR constructor
  """ __new__(cls: type,invalidHandleValue: IntPtr,ownsHandle: bool) """
  pass
 IsClosed=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the handle is closed.



Get: IsClosed(self: SafeHandle) -> bool



"""

 IsInvalid=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """When overridden in a derived class,gets a value indicating whether the handle value is invalid.



Get: IsInvalid(self: SafeHandle) -> bool



"""


 handle=None

